import copy
import os
import datetime
import hashlib
import re
import os.path
from Ganga.GPIDev.Base.Proxy import stripProxy, GPIProxyObjectFactory, isType, getName
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Lib.Job.Job import Job
from Ganga.Core.exceptions import GangaException
from Ganga.Utility.files import expandfilename
from GangaDirac.Lib.Utilities.DiracUtilities import getDiracEnv, execute
from Ganga.GPI import queues
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
configDirac = getConfig('DIRAC')
logger = getLogger()
regex = re.compile('[*?\[\]]')

global stored_list_of_sites
stored_list_of_sites = []


def all_SE_list(first_SE = ''):

    global stored_list_of_sites
    if stored_list_of_sites != []:
        return stored_list_of_sites

    all_storage_elements = configDirac['allDiracSE']
    if first_SE == '':
        default_SE = configDirac['DiracDefaultSE']
    else:
        default_SE = first_SE

    all_storage_elements.pop(all_storage_elements.index(default_SE))
    all_storage_elements.insert(0, default_SE)

    stored_list_of_sites = all_storage_elements

    return all_storage_elements

def getDefaultSE( ):
    return configDirac['DiracDefaultSE']

class DiracFile(IGangaFile):

    """
    File stored on a DIRAC storage element
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                                     'localDir': SimpleItem(defvalue=None, copyable=1, typelist=['str', 'type(None)'],
                                                            doc='local dir where the file is stored, used from get and put methods'),
                                     'locations': SimpleItem(defvalue=[], copyable=1, typelist=['str'], sequence=1,
                                                             doc="list of SE locations where the outputfiles are uploaded"),
                                     'compressed': SimpleItem(defvalue=False, typelist=['bool'], protected=0,
                                                              doc='wheather the output file should be compressed before sending somewhere'),
                                     'lfn': SimpleItem(defvalue='', copyable=1, typelist=['str'],
                                                       doc='return the logical file name/set the logical file name to use if not '
                                                       'using wildcards in namePattern'),
                                     'remoteDir': SimpleItem(defvalue="", doc='remote directory where the LFN is to be placed within '
                                                             'the dirac base directory by the put method.'),
                                     'guid': SimpleItem(defvalue='', copyable=1, typelist=['str'],
                                                        doc='return the GUID/set the GUID to use if not using wildcards in the namePattern.'),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], sequence=1, copyable=0,  # hidden=1,
                                                               typelist=['GangaDirac.Lib.Files.DiracFile'], doc="collected files from the wildcard namePattern"),
                                     'defaultSE': SimpleItem(defvalue='', copyable=1, doc="defaultSE where the file is to be accessed from or uploaded to"),
                                     'failureReason': SimpleItem(defvalue="", protected=1, copyable=0, doc='reason for the upload failure'),
    })

    _env = None

    _category = 'gangafiles'
    _name = "DiracFile"
    _exportmethods = ["get", "getMetadata", "getReplicas", 'getSubFiles', 'remove',
                      "replicate", 'put', 'locations', 'location', 'accessURL',
                      '_updateRemoteURLs', 'hasMatchedFiles']
    _remoteURLs = {}
    _storedReplicas = {}
    _have_copied = False

    def __init__(self, namePattern='', localDir=None, lfn='', remoteDir=None, **kwds):
        """
        name is the name of the output file that has to be written ...
        """

        super(DiracFile, self).__init__()
        self.locations = []

        if len(namePattern) >= 4 and str(namePattern).upper()[0:4] == "LFN:" and lfn == '':
            self._setLFNnamePattern(_lfn=namePattern[4:], _namePattern='')

        self._setLFNnamePattern(_lfn=lfn, _namePattern = namePattern)

        if localDir is not None:
            self.localDir = expandfilename(localDir)
        if remoteDir is not None:
            self.remoteDir = remoteDir

    def __construct__(self, args):

        self.locations = []

        if len(args) == 1 and isType(args[0], DiracFile):
            self.lfn = args[0].lfn
            self.namePattern = args[0].namePattern
            self.remoteDir = args[0].remoteDir
            self.localDir = args[0].localDir
            self.guid = args[0].guid
            self.compressed = args[0].compressed
            self._storedReplicas = args[0]._storedReplicas
            self._remoteURLs = args[0]._remoteURLs
            self.failureReason = args[0].failureReason
            self._have_copied = True
            self.subfiles = copy.deepcopy(args[0].subfiles)
            return

        # LFN ONLY
        if len(args) == 1 and type(args[0]) == type(''):
            if str(str(args[0]).upper()[0:4]) == str("LFN:"):
                self._setLFNnamePattern(_lfn=args[0][4:], _namePattern="")
            else:
                self._setLFNnamePattern(_lfn="", _namePattern=args[0])

        # NAMEPATTERN AND LFN
        elif len(args) == 2 and type(args[0]) == type('') and type(args[1]) == type(''):
            self.namePattern = args[0]
            self._setLFNnamePattern(_lfn='', _namePattern=self.namePattern)
            self.localDir = expandfilename(args[1])

        # NAMEPATTERN AND LFN AND LOCALDIR
        elif len(args) == 3 and type(args[0]) == type('') and type(args[1]) == type('') and type(args[2]) == type(''):
            self.namePattern = args[0]
            self.lfn = args[2]
            self._setLFNnamePattern(_lfn=self.lfn, _namePattern=self.namePattern)
            self.localDir = expandfilename(args[1])

        # NAMEPATTERN AND LFN AND LOCALDIR AND REMOTEDIR
        elif len(args) == 4 and type(args[0]) == type('') and type(args[1]) == type('')\
                and type(args[2]) == type('') and type(args[3]) == type(''):
            self.namePattern = args[0]
            self.lfn = args[2]
            self._setLFNnamePattern(_lfn=lfn, _namePattern=namePattern)
            self.localDir = expandfilename(args[1])
            self.remoteDir = args[3]

        # OTHER
        else:
            super(DiracFile, self).__construct__(args)

        return

    def __deepcopy__(self, memo):

        cls = type(stripProxy(self))
        c = super(cls, cls).__new__(cls)
        super(DiracFile, c).__init__()

        c.lfn = self.lfn
        c.localDir = self.lfn
        c.remoteDir = self.remoteDir
        c.namePattern = self.namePattern
        c.compressed = self.compressed
        c.locations = self.locations
        c.guid = self.guid
        c._storedReplicas = self._storedReplicas
        c._remoteURLs = self._remoteURLs
        c.failureReason = self.failureReason

        c.subfiles = copy.deepcopy(self.subfiles)
        c._have_copied = True
        return c

    def _attribute_filter__set__(self, name, value):

        if value != "" and value is not None:
            #   Do some checking of the filenames in a subprocess
            if name == 'lfn':
                if self.namePattern == '':
                    self.namePattern = os.path.basename(value)
                self.remoteDir = os.path.dirname(value)
                return value

            elif name == 'namePattern':
                self.localDir = os.path.dirname(value)
                return os.path.basename(value)

            elif name == 'localDir':
                return expandfilename(value)

        return value

    def locations(self):

        return self.locations

    def _setLFNnamePattern(self, _lfn="", _namePattern=""):

        if _lfn != "" and _lfn is not None:
            if len(_lfn) > 3 and _lfn[0:4] == "LFN:":
                _lfn = _lfn[4:]

        if _lfn != "" and _namePattern != "":
            self.lfn = _lfn
            self.remoteDir = os.path.dirname(_lfn)
            self.namePattern = os.path.basename(_namePattern)
            self.localDir = os.path.dirname(expandfilename(_namePattern))

        elif _lfn != "" and _namePattern == "":
            self.lfn = _lfn
            self.remoteDir = os.path.dirname(self.lfn)
            if self.namePattern != "":
                self.namePattern = os.path.basename(self.lfn)
            self.localDir = ""

        elif _namePattern != "" and _lfn == "":
            self.namePattern = os.path.basename(_namePattern)
            self.localDir = os.path.dirname(expandfilename(_namePattern))
            self.remoteDir = ""
            self.lfn = ""

    def _attribute_filter__get__(self, name):

        # Attempt to spend too long loading un-needed objects into memory in
        # order to read job status
        if name is 'lfn':
            if self.lfn == "":
                logger.warning("Do NOT have an LFN, for file: %s" % self.namePattern)
                logger.warning("If file exists locally try first using the method put()")
            return object.__getattribute__(self, 'lfn')
        elif name in ['guid', 'locations']:
            if configDirac['DiracFileAutoGet']:
                if name is 'guid':
                    if self.guid is None or self.guid == '':
                        if self.lfn != "":
                            self.getMetadata()
                            return object.__getattribute__(self, 'guid')
                if name is 'locations':
                    if self.locations == []:
                        if self.lfn != "":
                            self.getMetadata()
                            return object.__getattribute__(self, 'locations')

        return object.__getattribute__(self, name)

    def __repr__(self):
        """Get the representation of the file."""

        return "DiracFile(namePattern='%s', lfn='%s')" % (self.namePattern, self.lfn)

    def getSubFiles(self):
        """
        Returns the subfiles for this instance
        """
        if self.lfn == '':
            self.setLocation()
        return self.subfiles

    def dirac_line_processor(self, line, dirac_file, localPath):
        """
            Function to interperate the post processor lines.
            This returns False when everything went OK and True on an ERROR
        """
        logger.debug("Calling dirac_line_processor")
        tokens = line.strip().split(':::')
        logger.debug("dirac_line_processor: %s" % tokens)
        pattern = tokens[1].split('->')[0].split('&&')[0]
        name = tokens[1].split('->')[0].split('&&')[1]
        lfn = tokens[1].split('->')[1]
        guid = tokens[3]
        try:
            locations = eval(tokens[2])
        except:
            locations = tokens[2]

        if pattern == name:
            logger.debug("pattern == name")
            logger.error("Failed to parse outputfile data for file '%s'" % name)
            return True

        #   This is the case that multiple files were requested
        if pattern == dirac_file.namePattern:
            logger.debug("pattern == dirac_file.namePattern")
            d = DiracFile(namePattern=name, lfn=lfn)
            d.compressed = dirac_file.compressed
            d.guid = guid
            d.locations = locations
            d.localDir = localPath
            dirac_file.subfiles.append(d)
            #dirac_line_processor(line, d)
            return False

        #   This is the case that an individual file was requested
        elif name == dirac_file.namePattern:
            logger.debug("name == dirac_file.namePattern")
            if lfn == '###FAILED###':
                dirac_file.failureReason = tokens[2]
                logger.error("Failed to upload file '%s' to Dirac: %s" % (name, dirac_file.failureReason))
                return True
            dirac_file.lfn = lfn
            dirac_file.locations = locations
            dirac_file.guid = guid
            dirac_file.localDir = localPath
            return False

        else:
            return False

    def setLocation(self):
        """
        """

        logger.debug("DiracFile: setLocation")

        if not stripProxy(self).getJobObject():
            logger.error("No job assocaited with DiracFile: %s" % str(self))
            return

        job = self.getJobObject()
        postprocessLocationsPath = os.path.join(job.outputdir, getConfig('Output')['PostProcessLocationsFileName'])

        postprocesslocations = None

        try:
            postprocesslocations = open(postprocessLocationsPath, 'r')
            self.subfiles = []
            for line in postprocesslocations.readlines():
                if line.startswith('DiracFile'):
                    if self.dirac_line_processor(line, self, os.path.dirname(postprocessLocationsPath)) and regex.search(self.namePattern) is None:
                        logger.error("Error processing line:\n%s\nAND: namePattern: %s is NOT matched" % (str(line), str(self.namePattern)))

        except Exception, err:
            logger.warning("unexpected Error: %s" % str(err))
        finally:
            if postprocesslocations is not None:
                postprocesslocations.close()

    def _auto_remove(self):
        """
        Remove called when job is removed as long as config option allows
        """
        if self.lfn != '':
            queues.add(self.remove)

    def remove(self):
        """
        Remove this lfn and all replicas from DIRAC LFC/SEs
        """
        if self.lfn == "":
            raise GangaException(
                'Can\'t remove a  file from DIRAC SE without an LFN.')
        logger.info('Removing file %s' % self.lfn)
        stdout = execute('removeFile("%s")' % self.lfn)
        if isinstance(stdout, dict) and stdout.get('OK', False) and self.lfn in stdout.get('Value', {'Successful': {}})['Successful']:
            self.lfn = ""
            self.locations = []
            self.guid = ''
            return
        logger.error("Error in removing file '%s' : %s" % (self.lfn, stdout))
        return stdout

    def getMetadata(self):
        """
        Get Metadata associated with this files lfn. This method will also
        try to automatically set the files guid attribute.
        """

        if self.lfn == "":
            self._optionallyUploadLocalFile()

        # eval again here as datatime not included in dirac_ganga_server
        r = execute('getMetadata("%s")' % self.lfn)
        try:
            ret = eval(r)
        except:
            ret = r
        if isinstance(ret, dict) and ret.get('OK', False) and self.lfn in ret.get('Value', {'Successful': {}})['Successful']:
            try:
                if self.guid != ret['Value']['Successful'][self.lfn]['GUID']:
                    self.guid = ret['Value']['Successful'][self.lfn]['GUID']
            except:
                pass

        try:
            reps = self.getReplicas()
            ret['Value']['Successful'][self.lfn].update({'replicas': self.locations})
        except:
            pass

        return ret

    def _optionallyUploadLocalFile(self):
        """
        """

        if self.lfn != "":
            return

        if self.namePattern != "" and self.lfn == "":

            logger.info("I have a local DiracFile, however you're requesting it's location on the grid")
            logger.info("Shall I upload it to the grid before I continue?")
            decision = raw_input('[y] / n:')
            while not (decision in ['y', 'n'] or decision == ''):
                decision = raw_input('[y] / n:')

            if decision == 'y' or decision == '':
                # upload namePattern to grid
                logger.debug("Uploading the file first")
                self.put()
            elif decision == 'n':
                logger.debug("Not uploading now")
                return
            else:
                # do Nothing
                logger.debug("Continuing without uploading file")

            if self.lfn == "":
                raise GangaException('Uploading of namePattern: %s failed' % self.namePattern)

        if self.namePattern == "" and self.lfn == "":
            raise GangaException('Cannot do anything if I don\'t have an lfn or a namePattern!')

        return

    def getReplicas(self, forceRefresh=False):
        """
        Get the list of all SE where this file has a replica
        This relies on an internally stored list of replicas, (SE and  unless forceRefresh = True
        """

        if self.lfn == '':
            self._optionallyUploadLocalFile()
        if self.lfn == '':
            raise GangaException("Can't find replicas for file which has no LFN!")

        these_replicas = None

        if len(self.subfiles) != 0:

            allReplicas = []
            for i in self.subfiles:
                allReplicas.append(i.getReplicas())

            these_replicas = allReplicas

        else:
            # deep copy just before wer change it incase we're pointing to the
            # data stored in original from a copy
            if self._have_copied:
                self._storedReplicas = copy.deepcopy(self._storedReplicas)
            if (self._storedReplicas == {} and len(self.subfiles) == 0) or forceRefresh:

                self._storedReplicas = execute('getReplicas("%s")' % self.lfn)
                if self._storedReplicas.get('OK', False) is True:
                    try:
                        self._storedReplicas = self._storedReplicas['Value']['Successful']
                    except Exception as err:
                        logger.error("Unknown Error: %s" % str(err))
                        raise err
                else:
                    logger.error("Couldn't find replicas for: %s" % str(self.lfn))
                    raise GangaError("Couldn't find replicas for: %s" % str(self.lfn))
                logger.debug("getReplicas: %s" % str(self._storedReplicas))

                if self.lfn in self._storedReplicas.keys():
                    self._updateRemoteURLs(self._storedReplicas)

                    these_replicas = [self._storedReplicas[self.lfn]]
                else:
                    these_replicas = {}
            elif self._storedReplicas != {}:
                these_replicas = [self._storedReplicas[self.lfn]]

        return these_replicas

    def _updateRemoteURLs(self, reps):
        """
        Internal function used for storing all replica information about this LFN at different sites
        """
        if len(self.subfiles) != 0:
            for i in self.subfiles:
                i._updateRemoteURLs(reps)
        else:
            if self.lfn not in reps.keys():
                return
            if self.locations != reps[self.lfn].keys():
                self.locations = reps[self.lfn].keys()
            #logger.debug( "locations: %s" % str( self.locations ) )
            # deep copy just before wer change it incase we're pointing to the
            # data stored in original from a copy
            if self._have_copied:
                self._remoteURLs = copy.deepcopy(self._remoteURLs)
            for site in self.locations:
                #logger.debug( "site: %s" % str( site ) )
                self._remoteURLs[site] = reps[self.lfn][site]
                #logger.debug("Adding _remoteURLs[site]: %s" % str(self._remoteURLs[site]))

    def location(self):
        """
        Return a list of LFN locations for this DiracFile
        """
        if len(self.subfiles) == 0:
            if self.lfn == "":
                self._optionallyUploadLocalFile()
            else:
                return [self.lfn]
        else:
            # 1 LFN per DiracFile
            LFNS = []
            for this_file in self.subfiles:
                these_LFNs = this_file.location()
                for this_url in these_LFNs:
                    LFNs.append(this_url)
            return LFNs

    def accessURL(self, thisSE=''):
        """
        Attempt to find an accessURL which corresponds to an SE at this given site
        The given site will be provided either by thisSE which takes precedent or by
        the value stored in DiracFile.defaultSE
        """
        # If we don't have subfiles then we need to make sure that the replicas
        # are known
        if len(self._remoteURLs) == 0 and len(self.subfiles) == 0:
            self.getReplicas()

        # Now we have to match the replicas to find one at the
        # DiracDefaultSE
        if len(self.subfiles) == 0:

            files_URLs = self._remoteURLs
            this_accessURL = ''

            for this_SE in files_URLs.keys():

                this_URL = files_URLs.get(this_SE)
                these_sites_output = execute('getSitesForSE("%s")' % str(this_SE))
                if thisSE == '':
                    default_site = execute('getSiteForSE("%s")' % self.defaultSE)
                else:
                    default_site = thisSE

                if these_sites_output.get('OK', False):
                    these_sites = these_sites_output.get('Value')
                    for this_site in these_sites:
                        if type(default_site) == type([]):
                            hasMatched = False
                            for this_Site_in_SE in default_site:
                                if this_site == this_Site_in_SE:
                                    hasMatched = True
                                    break
                            if hasMatched:
                                break
                        elif type(default_site) == type(''):
                            if this_site == default_site:
                                this_accessURL = this_URL
                                break
                if this_accessURL != '':
                    break

            # Cannot find an accessURL
            if this_accessURL == '':
                return []

            # Currently only written to cope with 1 replica per DIRAC file
            # I think adding multiple accessURL for the same file at 1 site
            # adds confusion
            return [this_accessURL]
        else:
            # For all subfiles request the accessURL, 1 URL per LFN
            _accessURLs = []
            for i in self.subfiles:
                for j in i.accessURL():
                    _accessURLs.append(j)

            return _accessURLs

    def get(self, localPath=''):
        """
        Retrieves locally the file matching this DiracFile object pattern.
        If localPath is specified
        """
        if localPath == '':
            to_location = self.localDir

            if self.localDir is None:
                #to_location = os.getcwd()
                if self._parent is not None and os.path.isdir(self.getJobObject().outputdir):
                    to_location = self.getJobObject().outputdir
                else:
                    to_location = os.getcwd()
        else:
            to_location = localPath

        self.localDir = to_location

        if not os.path.isdir(to_location):
            raise GangaException(
                '"%s" is not a valid directory... Please set the localDir attribute to a valid directory' % self.localDir)

        if self.lfn == "":
            raise GangaException('Can\'t download a file without an LFN.')

        logger.info("Getting file %s" % self.lfn)
        stdout = execute('getFile("%s", destDir="%s")' % (self.lfn, to_location))
        if isinstance(stdout, dict) and stdout.get('OK', False) and self.lfn in stdout.get('Value', {'Successful': {}})['Successful']:
            if self.namePattern == "":
                name = os.path.basename(self.lfn)
                if self.compressed:
                    name = name[:-3]
                self.namePattern = name

            if self.guid == "" or not self.locations:
                self.getMetadata()
            return
        logger.error("Error in getting file '%s' : %s" % (self.lfn, str(stdout)))
        return stdout

    def replicate(self, destSE):
        """
        Replicate this file from self.locations[0] to destSE
        """
        if not self.locations:
            if self.lfn != '':
                self.getReplicas()
            else:
                raise GangaException('Can\'t replicate a file if it isn\'t already on a DIRAC SE, upload it first')
        if self.lfn == '':
            raise GangaException('Must supply an lfn to replicate')

        logger.info("Replicating file %s to %s" % (self.lfn, destSE))
        stdout = execute('replicateFile("%s", "%s", "%s")' % (self.lfn, destSE, self.locations[0]))
        if isinstance(stdout, dict) and stdout.get('OK', False) and self.lfn in stdout.get('Value', {'Successful': {}})['Successful']:
            self.locations.append(destSE)
            self.getReplicas(forceRefresh=True)
            return
        logger.error("Error in replicating file '%s' : %s" % (self.lfn, stdout))
        return stdout

    def processWildcardMatches(self):
        if regex.search(self.namePattern) is not None:
            raise Exception(
                "No wildcards in inputfiles for DiracFile just yet. Dirac are exposing this in API soon.")

    def put(self, lfn='', force=False, uploadSE="", replicate=False):
        """
        Try to upload file sequentially to storage elements defined in configDirac['allDiracSE'].
        File will be uploaded to the first SE that the upload command succeeds for.

        The file is uploaded to the SE described by the DiracFile.defaultSE attribute

        Alternatively, the user can specify an uploadSE which contains an SE
        which the file is to be uploaded to.

        If the user wants to replicate this file(s) across all SE then they should state replicate = True.

        Return value will be either the stdout from the dirac upload command if not
        using the wildcard characters '*?[]' in the namePattern.
        If the wildcard characters are used then the return value will be a list containing
        newly created DiracFile objects which were the result of glob-ing the wildcards.

        The objects in this list will have been uploaded or had their failureReason attribute populated if the
        upload failed.
        """

        if self.lfn != "" and force == False and lfn == '':
            logger.warning("Warning you're about to 'put' this DiracFile: %s on the grid as it already has an lfn: %s" % (self.namePattern, self.lfn))
            decision = raw_input('y / [n]:')
            while not (decision in ['y', 'n'] or decision == ''):
                decision = raw_input('y / [n]:')

            if decision == 'y':
                pass
            else:
                return

        if (lfn != '' and self.lfn != '') and force == False:
            logger.warning("Warning you're attempting to put this DiracFile: %s" % self.namePattern)
            logger.warning("It currently has an LFN associated with it: %s" % self.lfn)
            logger.warning("Do you want to continue and attempt to upload to: %s" % lfn)
            decision = raw_input('y / [n]:')
            while not (decision in ['y', 'n', '']):
                decision = raw_input('y / [n]:')

            if decision == 'y':
                pass
            else:
                return

        if lfn != '':
            self.lfn = lfn

        # looks like will only need this for the interactive uploading of jobs.
        # Also if any backend need dirac upload on client then when downloaded
        # this will upload then delete the file.

        if self.namePattern == "":
            if self.lfn != '':
                logger.warning("'Put'-ing a file with ONLY an existing LFN makes no sense!")
            raise GangaException('Can\'t upload a file without a local file name.')

        sourceDir = self.localDir
        if self.localDir is None:
            sourceDir = os.getcwd()
            # attached to a job, use the joboutputdir
            if self._parent != None and os.path.isdir(self.getJobObject().outputdir):
                sourceDir = self.getJobObject().outputdir

        if not os.path.isdir(sourceDir):
            raise GangaException('localDir attribute is not a valid dir, don\'t know from which dir to take the file')

        if regex.search(self.namePattern) is not None:
            if self.lfn != "":
                logger.warning("Cannot specify a single lfn for a wildcard namePattern")
                logger.warning("LFN will be generated automatically")
                self.lfn = ""

        selfConstructedLFN = False

        import glob
        if self.remoteDir == '' and self.lfn == '':
            import datetime
            t = datetime.datetime.now()
            this_date = t.strftime("%H.%M_%A_%d_%B_%Y")
            self.lfn = os.path.join(configDirac['DiracLFNBase'], 'GangaFiles_%s' % this_date)
            selfConstructedLFN = True
        #if self.remoteDir == '' and self.lfn != '':
        #    self.remoteDir = configDirac['DiracLFNBase']

        if self.remoteDir[:4] == 'LFN:':
            lfn_base = self.remoteDir[4:]
        else:
            lfn_base = self.remoteDir

        if uploadSE != "":
            storage_elements = all_SE_list(uploadSE)
        else:
            storage_elements = all_SE_list(self.defaultSE)

        outputFiles = GangaList()
        for this_file in glob.glob(os.path.join(sourceDir, self.namePattern)):
            name = this_file

            if not os.path.exists(name):
                if not self.compressed:
                    raise GangaException('Cannot upload file. File "%s" must exist!' % name)
                name += '.gz'
                if not os.path.exists(name):
                    raise GangaException('File "%s" must exist!' % name)
            else:
                if self.compressed:
                    os.system('gzip -c %s > %s.gz' % (name, name))
                    name += '.gz'
                    if not os.path.exists(name):
                        raise GangaException('File "%s" must exist!' % name)

            if lfn == "":
                lfn = os.path.join(lfn_base, os.path.basename(name))

            if selfConstructedLFN is True:
                self.lfn = os.path.join(self.lfn, os.path.basename(name))

            lfn = self.lfn

            d = DiracFile()
            d.namePattern = os.path.basename(name)
            d.compressed = self.compressed
            d.localDir = sourceDir
            stderr = ''
            stdout = ''
            logger.info('Uploading file %s to %s as %s' % (name, storage_elements[0], lfn))
            stdout = execute('uploadFile("%s", "%s", %s)' % (lfn, name, str([storage_elements[0]])))
            if type(stdout) == str:
                logger.warning("Couldn't upload file '%s': %s" % (os.path.basename(name), stdout))
                continue
            if stdout.get('OK', False) and lfn in stdout.get('Value', {'Successful': {}})['Successful']:
                # when doing the two step upload delete the temp file
                if self.compressed or self._parent != None:
                    os.remove(name)
                # need another eval as datetime needs to be included.
                guid = stdout['Value']['Successful'][lfn].get('GUID', '')
                if regex.search(self.namePattern) is not None:
                    d.lfn = lfn
                    d.remoteDir = os.path.dirname(lfn)
                    d.locations = stdout['Value']['Successful'][lfn].get('allDiracSE', '')
                    d.guid = guid
                    outputFiles.append(GPIProxyObjectFactory(d))
                    continue
                else:
                    self.lfn = lfn
                    self.remoteDir = os.path.dirname(lfn)
                    self.locations = stdout['Value']['Successful'][lfn].get('allDiracSE', '')
                    self.guid = guid
                # return ## WHY?
            else:
                failureReason = "Error in uploading file %s : %s" % (os.path.basename(name), str(stdout))
                logger.error(failureReason)
                if regex.search(self.namePattern) is not None:
                    d.failureReason = failureReason
                    outputFiles.append(GPIProxyObjectFactory(d))
                    continue
                self.failureReason = failureReason
                return str(stdout)

        if replicate == True:

            if len(outputFiles) == 1 or len(outputFiles) == 0:
                storage_elements.pop(0)
                for se in storage_elements:
                    self.replicate(se)
            else:
                storage_elements.pop(0)
                for this_file in outputFiles:
                    for se in storage_elements:
                        this_file.replicate(se)

        if len(outputFiles) > 0:
            return GPIProxyObjectFactory(outputFiles)
        else:
            outputFiles.append(self)
            return GPIProxyObjectFactory(outputFiles)

    def getWNScriptDownloadCommand(self, indent):

        import inspect
        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), 'downloadScript.py')

        from Ganga.GPIDev.Lib.File import FileUtils
        download_script = FileUtils.loadScript(script_location, '')

        script = """\n
download_script='''\n###DOWNLOAD_SCRIPT###'''
import subprocess
dirac_env=###DIRAC_ENV###
subprocess.Popen('''python -c "import sys\nexec(sys.stdin.read())"''', shell=True, env=dirac_env, stdin=subprocess.PIPE).communicate(download_script)
"""
        script = '\n'.join([ str(indent+str(line)) for line in script.split('\n')])

        replace_dict = {'###DOWNLOAD_SCRIPT###' : download_script,
                        '###DIRAC_ENV###' : self._getDiracEnvStr(),
                        '###LFN###' : self.lfn}

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def _getDiracEnvStr(self):
        from GangaDirac.Lib.Utilities.DiracUtilities import getDiracEnv
        diracEnv = str(getDiracEnv())
        return diracEnv

    def _WN_wildcard_script(self, namePattern, lfnBase, compressed):
        wildcard_str =  """
import glob, hashlib
for f in glob.glob('###NAME_PATTERN###'):
    processes.append(uploadFile(os.path.basename(f), '###LFN_BASE###', ###COMPRESSED###, '###NAME_PATTERN###'))
"""
        from Ganga.GPIDev.Lib.File import FileUtils
        wildcard_str = FileUtils.indentScript(wildcard_str, '###INDENT###')

        replace_dict = { '###NAME_PATTERN###' : namePattern,
                         '###LFN_BASE###' : lfnBase,
                         '###COMPRESSED###' : compressed }

        for k, v in replace_dict.iteritems():
            wildcard_str = wildcard_str.replace(str(k), str(v))

        return wildcard_str

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """

        import inspect
        script_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        script_location = os.path.join( script_path, 'uploadScript.py')

        from Ganga.GPIDev.Lib.File import FileUtils
        upload_script = FileUtils.loadScript(script_location, '')

        WNscript_location = os.path.join( script_path, 'WNInjectTemplate.py' )
        script = FileUtils.loadScript(WNscript_location, '###INDENT###')

        selfConstructedLFNs = False

        if self.remoteDir == '' and self.lfn == '':
            import datetime
            t = datetime.datetime.now()
            this_date = t.strftime("%H.%M_%A_%d_%B_%Y")
            self.lfn = os.path.join(configDirac['DiracLFNBase'], 'GangaFiles_%s' % this_date)
            selfConstructedLFNs = True

        if self.remoteDir == '' and self.lfn != '':
            self.remoteDir = configDirac['DiracLFNBase']

        if self.remoteDir[:4] == 'LFN:':
            lfn_base = self.remoteDir[4:]
        else:
            lfn_base = self.remoteDir


        for this_file in outputFiles:
            isCompressed = this_file.namePattern in patternsToZip

            if not regex.search(this_file.namePattern) is None:
                script += self._WN_wildcard_script(this_file.namePattern, lfn_base, str(isCompressed))
            else:
                script += '###INDENT###print("Uploading: %s as: %s")\n' % (this_file.namePattern, str(os.path.join(lfn_base, this_file.namePattern)))
                script += '###INDENT###processes.append(uploadFile("%s", "%s", %s))\n' % (this_file.namePattern, lfn_base, str(isCompressed))


        if stripProxy(self)._parent is not None and stripProxy(self).getJobObject() and getName(stripProxy(self).getJobObject().backend) != 'Dirac':
            script_env = self._getDiracEnvStr()
        else:
            script_env = str(None)

        script = '\n'.join([str('###INDENT###' + str(line)) for line in script.split('\n')])

        replace_dict = {'###UPLOAD_SCRIPT###' : upload_script,
                        '###STORAGE_ELEMENTS###' : str(configDirac['allDiracSE']),
                        '###INDENT###' : indent,
                        '###LOCATIONSFILE###' : postProcessLocationsFP,
                        '###DIRAC_ENV###' : script_env}

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def hasMatchedFiles(self):

        if self.lfn != "" and self.namePattern != "":
            if self.namePattern == os.path.basename(self.lfn):
                return True
            else:
                logger.error("LFN doesn't match namePattern for file: %s" % str(self.namePattern))
                return False
        elif len(self.subfiles) > 0 and regex.search(self.namePattern) is not None:
            return True
        else:
            logger.error("Failed to Match file:\n%s" % str(self))
            return False

# add DiracFile objects to the configuration scope (i.e. it will be
# possible to write instatiate DiracFile() objects via config file)
import Ganga.Utility.Config
Ganga.Utility.Config.config_scope['DiracFile'] = DiracFile

from Ganga.Runtime.GPIexport import exportToGPI
exportToGPI('GangaDirac', GangaList, 'Classes')




###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ExecutablePandaRTHandler.py,v 1.4 2009/04/22 07:43:44 dvanders Exp $
###############################################################################
# Athena LCG Runtime Handler
#
# ATLAS/ARDA

import os, sys, pwd, commands, re, shutil, urllib, time, string, exceptions, time

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.Core import BackendError
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *

from GangaAtlas.Lib.ATLASDataset import DQ2Dataset, DQ2OutputDataset
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2outputdatasetname
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_set_dataset_lifetime

from GangaPanda.Lib.Panda.Panda import setChirpVariables, uploadSources

class ExecutablePandaRTHandler(IRuntimeHandler):
    '''Executable Panda Runtime Handler'''

    def master_prepare(self,app,appconfig):
        '''Prepare the master job'''

        from pandatools import Client
        from taskbuffer.JobSpec import JobSpec
        from taskbuffer.FileSpec import FileSpec

        job = app._getParent()
        logger.debug('ExecutablePandaRTHandler master_prepare called for %s', job.getFQID('.')) 

        # set chirp variables
        if configPanda['chirpconfig'] or configPanda['chirpserver']:
            setChirpVariables()

#       Pack inputsandbox
        inputsandbox = 'sources.%s.tar' % commands.getoutput('uuidgen 2> /dev/null') 
        inpw = job.getInputWorkspace()
        # add user script to inputsandbox
        if hasattr(job.application.exe, "name"):
            if not job.application.exe in job.inputsandbox:
                job.inputsandbox.append(job.application.exe)

        for fname in [f.name for f in job.inputsandbox]:
            fname.rstrip(os.sep)
            path = fname[:fname.rfind(os.sep)]
            f = fname[fname.rfind(os.sep)+1:]
            rc, output = commands.getstatusoutput('tar rf %s -C %s %s' % (inpw.getPath(inputsandbox), path, f))
            if rc:
                logger.error('Packing inputsandbox failed with status %d',rc)
                logger.error(output)
                raise ApplicationConfigurationError(None,'Packing inputsandbox failed.')
        if len(job.inputsandbox) > 0:
            rc, output = commands.getstatusoutput('gzip %s' % (inpw.getPath(inputsandbox)))
            if rc:
                logger.error('Packing inputsandbox failed with status %d',rc)
                logger.error(output)
                raise ApplicationConfigurationError(None,'Packing inputsandbox failed.')
            inputsandbox += ".gz"
        else:
            inputsandbox = None

#       Upload Inputsandbox
        if inputsandbox:
            logger.debug('Uploading source tarball ...')
            uploadSources(inpw.getPath(),os.path.basename(inputsandbox))
            self.inputsandbox = inputsandbox
        else:
            self.inputsandbox = None

#       input dataset
        if job.inputdata:
            if job.inputdata._name != 'DQ2Dataset':
                raise ApplicationConfigurationError(None,'PANDA application supports only DQ2Datasets')

        # run brokerage here if not splitting
        if not job.splitter:
            from GangaPanda.Lib.Panda.Panda import runPandaBrokerage
            runPandaBrokerage(job)
        elif job.splitter._name not in ['DQ2JobSplitter', 'ArgSplitter', 'ArgSplitterTask']:
            raise ApplicationConfigurationError(None,'Panda splitter must be DQ2JobSplitter or ArgSplitter')
        
        if job.backend.site == 'AUTO':
            raise ApplicationConfigurationError(None,'site is still AUTO after brokerage!')

#       output dataset
        if job.outputdata:
            if job.outputdata._name != 'DQ2OutputDataset':
                raise ApplicationConfigurationError(None,'Panda backend supports only DQ2OutputDataset')
        else:
            logger.info('Adding missing DQ2OutputDataset')
            job.outputdata = DQ2OutputDataset()

        job.outputdata.datasetname,outlfn = dq2outputdatasetname(job.outputdata.datasetname, job.id, job.outputdata.isGroupDS, job.outputdata.groupname)

        self.outDsLocation = Client.PandaSites[job.backend.site]['ddm']

        try:
            Client.addDataset(job.outputdata.datasetname,False,location=self.outDsLocation)
            logger.info('Output dataset %s registered at %s'%(job.outputdata.datasetname,self.outDsLocation))
            dq2_set_dataset_lifetime(job.outputdata.datasetname, location=self.outDsLocation)
        except exceptions.SystemExit:
            raise BackendError('Panda','Exception in Client.addDataset %s: %s %s'%(job.outputdata.datasetname,sys.exc_info()[0],sys.exc_info()[1]))

        # handle the libds
        if job.backend.libds:
            self.libDataset = job.backend.libds
            self.fileBO = getLibFileSpecFromLibDS(self.libDataset)
            self.library = self.fileBO.lfn
        elif job.backend.bexec:
            self.libDataset = job.outputdata.datasetname+'.lib'
            self.library = '%s.tgz' % self.libDataset
            try:
                Client.addDataset(self.libDataset,False,location=self.outDsLocation)
                dq2_set_dataset_lifetime(self.libDataset, location=self.outDsLocation)
                logger.info('Lib dataset %s registered at %s'%(self.libDataset,self.outDsLocation))
            except exceptions.SystemExit:
                raise BackendError('Panda','Exception in Client.addDataset %s: %s %s'%(self.libDataset,sys.exc_info()[0],sys.exc_info()[1]))

        # collect extOutFiles
        self.extOutFile = []
        for tmpName in job.outputdata.outputdata:
            if tmpName != '':
                self.extOutFile.append(tmpName)

        for tmpName in job.outputsandbox:
            if tmpName != '':
                self.extOutFile.append(tmpName)

        for tmpName in job.backend.extOutFile:
            if tmpName != '':
                self.extOutFile.append(tmpName)

        # create build job
        if job.backend.bexec != '':
            jspec = JobSpec()
            jspec.jobDefinitionID   = job.id
            jspec.jobName           = commands.getoutput('uuidgen 2> /dev/null')
            jspec.transformation    = '%s/buildGen-00-00-01' % Client.baseURLSUB
            if Client.isDQ2free(job.backend.site):
                jspec.destinationDBlock = '%s/%s' % (job.outputdata.datasetname,self.libDataset)
                jspec.destinationSE     = 'local'
            else:
                jspec.destinationDBlock = self.libDataset
                jspec.destinationSE     = job.backend.site
            jspec.prodSourceLabel   = configPanda['prodSourceLabelBuild']
            jspec.processingType    = configPanda['processingType']
            jspec.assignedPriority  = configPanda['assignedPriorityBuild']
            jspec.computingSite     = job.backend.site
            jspec.cloud             = job.backend.requirements.cloud
            jspec.jobParameters     = '-o %s' % (self.library)
            if self.inputsandbox:
                jspec.jobParameters     += ' -i %s' % (self.inputsandbox)
            else:
                raise ApplicationConfigurationError(None,'Executable on Panda with build job defined, but inputsandbox is emtpy !')
            matchURL = re.search('(http.*://[^/]+)/',Client.baseURLCSRVSSL)
            if matchURL:
                jspec.jobParameters += ' --sourceURL %s ' % matchURL.group(1)
            if job.backend.bexec != '':
                jspec.jobParameters += ' --bexec "%s" ' % urllib.quote(job.backend.bexec)
                jspec.jobParameters += ' -r %s ' % '.'
                

            fout = FileSpec()
            fout.lfn  = self.library
            fout.type = 'output'
            fout.dataset = self.libDataset
            fout.destinationDBlock = self.libDataset
            jspec.addFile(fout)

            flog = FileSpec()
            flog.lfn = '%s.log.tgz' % self.libDataset
            flog.type = 'log'
            flog.dataset = self.libDataset
            flog.destinationDBlock = self.libDataset
            jspec.addFile(flog)
            return jspec
        else:
            return None

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        '''prepare the subjob specific configuration'''
 
        from pandatools import Client
        from taskbuffer.JobSpec import JobSpec
        from taskbuffer.FileSpec import FileSpec

        job = app._getParent()
        logger.debug('AthenaPandaRTHandler prepare called for %s', job.getFQID('.'))

#       in case of a simple job get the dataset content, otherwise subjobs are filled by the splitter
        if job.inputdata and not job._getRoot().subjobs:

            if not job.inputdata.names:
                
                contents = job.inputdata.get_contents(overlap=False, size=True)

                for ds in contents.keys():

                    for f in contents[ds]:
                        job.inputdata.guids.append( f[0] )
                        job.inputdata.names.append( f[1][0] )
                        job.inputdata.sizes.append( f[1][1] )
                        job.inputdata.checksums.append( f[1][2] )
                        job.inputdata.scopes.append( f[1][3] )


        site = job._getRoot().backend.site
        job.backend.site = site
        job.backend.actualCE = site
        cloud = job._getRoot().backend.requirements.cloud
        job.backend.requirements.cloud = cloud

#       if no outputdata are given
        if not job.outputdata:
            job.outputdata = DQ2OutputDataset()
            job.outputdata.datasetname = job._getRoot().outputdata.datasetname
        #if not job.outputdata.datasetname:
        else:
            job.outputdata.datasetname = job._getRoot().outputdata.datasetname

        if not job.outputdata.datasetname:
            raise ApplicationConfigurationError(None,'DQ2OutputDataset has no datasetname')

        jspec = JobSpec()
        jspec.jobDefinitionID   = job._getRoot().id
        jspec.jobName           = commands.getoutput('uuidgen 2> /dev/null')
        jspec.transformation    = '%s/runGen-00-00-02' % Client.baseURLSUB
        if job.inputdata:
            jspec.prodDBlock    = job.inputdata.dataset[0]
        else:
            jspec.prodDBlock    = 'NULL'
        jspec.destinationDBlock = job.outputdata.datasetname
        if job.outputdata.location:
            if not job._getRoot().subjobs or job.id == 0:
                logger.warning('You have specified outputdata.location. Note that Panda may not support writing to a user-defined output location.')
            jspec.destinationSE = job.outputdata.location
        else:
            jspec.destinationSE = site
        jspec.prodSourceLabel   = configPanda['prodSourceLabelRun']
        jspec.processingType    = configPanda['processingType']
        jspec.assignedPriority  = configPanda['assignedPriorityRun']
        jspec.cloud             = cloud
        # memory
        if job.backend.requirements.memory != -1:
            jspec.minRamCount = job.backend.requirements.memory
        # cputime     
        if job.backend.requirements.cputime != -1:
            jspec.maxCpuCount = job.backend.requirements.cputime
        jspec.computingSite     = site

#       library (source files)
        if job.backend.libds:
            flib = FileSpec()
            flib.lfn            = self.fileBO.lfn
            flib.GUID           = self.fileBO.GUID
            flib.type           = 'input'
            flib.status         = self.fileBO.status
            flib.dataset        = self.fileBO.destinationDBlock
            flib.dispatchDBlock = self.fileBO.destinationDBlock
            jspec.addFile(flib)
        elif job.backend.bexec:
            flib = FileSpec()
            flib.lfn            = self.library
            flib.type           = 'input'
            flib.dataset        = self.libDataset
            flib.dispatchDBlock = self.libDataset
            jspec.addFile(flib)

#       input files FIXME: many more input types
        if job.inputdata:            
            for guid, lfn, size, checksum, scope in zip(job.inputdata.guids,job.inputdata.names,job.inputdata.sizes, job.inputdata.checksums, job.inputdata.scopes):
                finp = FileSpec()
                finp.lfn            = lfn
                finp.GUID           = guid
                finp.scope          = scope
                
#            finp.fsize =
#            finp.md5sum =
                finp.dataset        = job.inputdata.dataset[0]
                finp.prodDBlock     = job.inputdata.dataset[0]
                finp.dispatchDBlock = job.inputdata.dataset[0]
                finp.type           = 'input'
                finp.status         = 'ready'
                jspec.addFile(finp)

#       output files
#        outMap = {}
        
        #FIXME: if options.outMeta != []:
        self.rundirectory = "."

#       log files

        flog = FileSpec()
        flog.lfn = '%s._$PANDAID.log.tgz' % job.outputdata.datasetname
        flog.type = 'log'
        flog.dataset           = job.outputdata.datasetname
        flog.destinationDBlock = job.outputdata.datasetname
        flog.destinationSE     = job.backend.site
        jspec.addFile(flog)

#       job parameters
        param = ''

        # source URL
        matchURL = re.search("(http.*://[^/]+)/",Client.baseURLCSRVSSL)
        srcURL = ""
        if matchURL != None:
            srcURL = matchURL.group(1)
            param += " --sourceURL %s " % srcURL

        param += '-r "%s" ' % self.rundirectory

        exe_name = job.application.exe
        if job.backend.bexec == '':
            if hasattr(job.application.exe, "name"):
                exe_name = os.path.basename(job.application.exe.name)

            # set jobO parameter
            if job.application.args:
                param += ' -j "" -p "%s %s" '%(exe_name,urllib.quote(" ".join(job.application.args)))
            else:
                param += ' -j "" -p "%s" '%exe_name
            if self.inputsandbox:
                param += ' -a %s '%self.inputsandbox

        else:
            param += '-l %s ' % self.library
            param += '-j "" -p "%s %s" ' % ( exe_name,urllib.quote(" ".join(job.application.args)))

        if job.inputdata:
            param += '-i "%s" ' % job.inputdata.names

        # fill outfiles
        outfiles = {}
        for f in self.extOutFile:
            tarnum = 1
            if f.find('*') != -1:
            # archive *
                outfiles[f] = "outputbox%i.%s.%s.tar.gz" % (tarnum, job.getFQID('.'), time.strftime("%Y%m%d%H%M%S") )
                tarnum += 1
            else:
                outfiles[f] = "%s.%s.%s" %(f, job.getFQID('.'), time.strftime("%Y%m%d%H%M%S"))

            fout = FileSpec()
            fout.lfn = outfiles[f]
            fout.type = 'output'
            fout.dataset           = job.outputdata.datasetname
            fout.destinationDBlock = job.outputdata.datasetname
            fout.destinationSE     = job.backend.site
            jspec.addFile(fout)

        param += '-o "%s" ' % (outfiles) # must be double quotes, because python prints strings in 'single quotes' 

        for file in jspec.Files:
            if file.type in [ 'output', 'log'] and configPanda['chirpconfig']:
                file.dispatchDBlockToken = configPanda['chirpconfig']
                logger.debug('chirp file %s',file)

        jspec.jobParameters = param
        
        return jspec

from Ganga.GPIDev.Credentials import GridProxy
gridProxy = GridProxy()

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('Executable','Panda',ExecutablePandaRTHandler)

from Ganga.Utility.Config import getConfig, ConfigError
config = getConfig('Athena')
configDQ2 = getConfig('DQ2')
configPanda = getConfig('Panda')

from Ganga.Utility.logging import getLogger
logger = getLogger()



from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks import Task
from MCTransforms import EvgenTransform, SimulTransform, ReconTransform
from GangaAtlas.Lib.AthenaMC.AthenaMCDatasets import AthenaMCInputDatasets
from Ganga.GPIDev.Schema import *
from Ganga.Utility.ColourText import ANSIMarkup, fgcol, Effects
markup = ANSIMarkup()
fx = Effects()


from Ganga.GPIDev.Base.Objects import Node

o = ""
o+="\n"+ markup(" *** Task for Athena Monte Carlo production ***", fgcol("blue"))
def c(s):
   return markup(s,fgcol("blue"))
o+="\n"+ " Set the total number of events to be processed: "+c("t.total_events = 10000")
o+="\n"+ " Applications for evgen/simul/recon            : "+c("t.evgen.application, t.simul.application, t.recon.application")
o+="\n"+ " Events processed per evgen/simul/recon        : "+c("t.evgen.application.number_events_job = 10000")
o+="\n"+ " Skip events / files from the previous dataset : "+c("t.simul.inputdata.skip_events = 1000")
o+="\n"+ " Use an input dataset for the event generator  : "+c("t.initializeFromGenerator(dataset, events_per_file_in_dataset)")
o+="\n"+ " Use an existing evgen EVNT dataset            : "+c("t.initializeFromEvgen(dataset, events_per_file_in_dataset)")
o+="\n"+ " Use an existing simul RDO dataset             : "+c("t.initializeFromSimul(dataset, events_per_file_in_dataset)")
o+="\n"
o+="\n"+ markup("Procedure to do a usual production:", fgcol("red"))
o+="\n"+ "t = MCTask()"
o+="\n"+ 't.name = "MyFirstTask"'
o+="\n"+ 't.evgen.application.evgen_job_option = "CSC.005145.PythiaZmumu.py" # or "path/to/my_evgen_job_option.py"'
o+="\n"+ 't.evgen.application.transform_script = "csc_evgen08_trf.py"'
o+="\n"+ "t.total_events = 1000"
o+="\n"+ 't.setParameter(run_number="5145", production_name="MyProd-01", process_name="PythiaZmumu")'
o+="\n"+ 't.setParameter(atlas_release="14.2.24.1", se_name="FZK-LCG2_SCRATCHDISK")'
o+="\n"+ 't.setParameter(triggerConfig="DEFAULT", geometryTag="ATLAS-GEO-05-00-00")'
o+="\n"+ "t.float = 10"
o+="\n"+ "t.info() # Check here if settings are correct"
o+="\n"+ "t.run()"
o+="\n"+ "t.overview() # Watch the processing"
mctask_help = o
mctask_help_nocolor = mctask_help.replace(fgcol("blue"),"").replace(fx.normal, "").replace(fgcol("red"),"")

class MCTask(Task):
   __doc__ = mctask_help_nocolor
   _schema = Schema(Version(1,1), dict(Task._schema.datadict.items() + {
        'total_events': SimpleItem(defvalue=1000, doc="Total number of events to generate", typelist=["int"]),
        'evgen': SimpleItem(defvalue=None, transient=1, typelist=["object"], doc='Evgen Transform'),
        'simul': SimpleItem(defvalue=None, transient=1, typelist=["object"], doc='Simul Transform'),
        'recon': SimpleItem(defvalue=None, transient=1, typelist=["object"], doc='Recon Transform'),
       }.items()))
   _category = 'tasks'
   _name = 'MCTask'
   _exportmethods = Task._exportmethods + [
                    'initializeFromGenerator', 'initializeFromEvgen', 'initializeFromSimul',
                    ]

## Special methods
   def initialize(self):
      super(MCTask,self).initialize()
      evgen = EvgenTransform()
      simul = SimulTransform()
      recon = ReconTransform()
      self.transforms = [evgen,simul,recon]
      self.setBackend(GPI.LCG())

   def startup(self):
      super(MCTask,self).startup()
      self.initAliases()

   def check(self):
      self.initAliases()
      # Set the status in reverse order so the propagation works correctly
      # the minus signs are used to get a correct rounding behaviour
      for i in range(1,len(self.transforms)):
         if self.transforms[i].inputdata:
            self.transforms[i].inputdata.use_partition_numbers = True
            self.transforms[i].inputdata.number_events_file = self.transforms[i-1].application.number_events_job
      for i in range(len(self.transforms)-1,-1,-1):
         tf = self.transforms[i]
         lastpartition = -((-self.total_events)/tf.application.number_events_job)
         tf.setPartitionsLimit(lastpartition+1)
         tf.setPartitionsStatus([c for c in range(1,lastpartition+1) if tf.getPartitionStatus(c) == "ignored"], "ready")
      # if the first transformation has an input dataset, check which partitions are ready
      if len(self.transforms) > 0 and self.transforms[0].inputdata:
         tf = self.transforms[0]
         tf.inputdata.get_dataset(tf.application, tf.backend._name)
         inputnumbers = tf.inputdata.filesToNumbers(tf.inputdata.turls.keys())
         partitions = tf.application.getPartitionsForInputs(inputnumbers, tf.inputdata)
         tf.setPartitionsStatus([c for c in range(1,lastpartition+1) if c in partitions], "ready")
         tf.setPartitionsStatus([c for c in range(1,lastpartition+1) if not c in partitions], "hold")
      super(MCTask,self).check()

   def initAliases(self):
      self.evgen = None
      self.simul = None
      self.recon = None
      for tf in self.transforms:
         if "Recon" in tf.__class__.__name__:
            self.recon = tf
         elif "Simul" in tf.__class__.__name__:
            self.simul = tf
         elif "Evgen" in tf.__class__.__name__:
            self.evgen = tf

## Public methods
   def initializeFromGenerator(self,dataset,events_per_file):
      self.initialize()
      self.transforms[0].inputdata = AthenaMCInputDatasets()
      self.transforms[0].inputdata.use_partition_numbers = True
      self.transforms[0].inputdata.DQ2dataset = dataset
      self.transforms[0].inputdata.number_events_file = events_per_file
      self.check()

   def initializeFromEvgen(self,dataset,events_per_file):
      self.initialize()
      self.removeTransform(0) # remove evgen
      self.transforms[0].inputdata.use_partition_numbers = True
      self.transforms[0].inputdata.DQ2dataset = dataset
      self.transforms[0].inputdata.number_events_file = events_per_file
      self.check()

   def initializeFromSimul(self,dataset,events_per_file):
      self.initialize()
      self.removeTransform(0) # remove evgen
      self.removeTransform(0) # remove simul
      self.transforms[0].inputdata.use_partition_numbers = True
      self.transforms[0].inputdata.DQ2dataset = dataset
      self.transforms[0].inputdata.number_events_file = events_per_file
      self.check()

## Information methods
   def info(self):
      print "* total events: %i\n" % (self.total_events) 
      super(MCTask,self).info()

   def help(self):
      print mctask_help




################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 16/12/2013
################################################################################

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IPrepareApp import IPrepareApp
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config import getConfig

from Ganga.GPIDev.Lib.File import *
from Ganga.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.Core import ApplicationConfigurationError

import os, shutil, commands, re, time
from Ganga.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class VFT_make_ana(IApplication):
    """
    VFT_make_ana application running reconUtils/macros/grtf_VFT/make_ana.py

    TODO documentation:
    - It is possible to run make_ana alone or make_ana+make_pdf
    - Explain that the make_ana and make_pdf arguments become respectively
      ana_xxxx and pdf_xxxx in this application's setup arguments
    
    """
    _schema = Schema(Version(1,1), {
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'tree' : SimpleItem(defvalue=None,doc='Name of the TTree to analyze in the input file(s). The default name is ND280ValidTree.', typelist=['str','type(None)']),
        'ana_custom' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'ana_output' : SimpleItem(defvalue=None,doc='Output filename or filenames. Takes a string or a list of strings.', typelist=['str','type(None)']),
        'ana_useropt' : SimpleItem(defvalue=[],typelist=['str'],sequence=1,strict_sequence=0,doc="List of user options to pass to make_ana.py using -0. No need to provide '-O'."),

        'run_pdf' : SimpleItem(defvalue=True,doc='Switch to run make_pdf.py on the make_ana.py output.', typelist=['bool']),
        'pdf_custom' : SimpleItem(defvalue=None,doc='Name of the make_pdf.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_title' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_rdp' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_mcp' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_oldrdp' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_oldmcp' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_rdptitle' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_mcptitle' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_oldrdptitle' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_oldmcptitle' : SimpleItem(defvalue=None,doc='Name of the make_ana.py module to use for this analysis.', typelist=['str','type(None)']),
        'pdf_output' : SimpleItem(defvalue=None,doc='Output filename or filenames. Takes a string or a list of strings.', typelist=['str','type(None)']),
        'pdf_options' : SimpleItem(defvalue=[],typelist=['str'],sequence=1,strict_sequence=0,doc="List of user options to pass to make_pdf.py."),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        } )
    _category = 'applications'
    _name = 'VFT_make_ana'
    _scriptname = None
    _exportmethods = []
    _GUIPrefs = [ { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                  { 'attribute' : 'tree', 'widget' : 'String' },
                  { 'attribute' : 'ana_custom', 'widget' : 'String' },
                  { 'attribute' : 'ana_output', 'widget' : 'String' },
                  { 'attribute' : 'ana_useropt', 'widget' : 'String_List' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ 
                          { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                          { 'attribute' : 'tree', 'widget' : 'String' },
                          { 'attribute' : 'ana_custom', 'widget' : 'String' },
                          { 'attribute' : 'ana_output', 'widget' : 'String' },
                          { 'attribute' : 'ana_useropt', 'widget' : 'String_List' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]


    def __init__(self):
        super(VFT_make_ana,self).__init__()


    def configure(self,masterappconfig):
        
        self.ana_useropt = convertIntToStringArgs(self.ana_useropt)
        args = []
        args.append('$RECONUTILSROOT/macros/grtf_VFT/make_ana.py')

        job = self.getJobObject()

        if self.cmtsetup == None:
          raise ApplicationConfigurationError(None,'No cmt setup script given.')

        if not self.tree == None:
          args.append('-t')
          args.append(self.tree)

        if not self.ana_custom == None:
          args.append('-c')
          args.append(self.ana_custom)

        if not self.ana_useropt == None:
          for UsrOpt in self.ana_useropt:
            args.append('-O')
            args.append(UsrOpt)

        if self.ana_output == None:
          raise ApplicationConfigurationError(None,'No output file given. Fill the ana_output variable.')
        else:
          args.append('-o')
          args.append(self.ana_output)


        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        if job.inputdata == None:
          raise ApplicationConfigurationError(None,'The inputdata variable is not defined.')
        fileList = job.inputdata.get_dataset_filenames()
        if len(fileList) < 1:
          raise ApplicationConfigurationError(None,'No input data file given.')
        args.extend(fileList)

        if self.run_pdf:
          args.append('&&')
          args.append('$ND280ANALYSISTOOLSROOT/macros/grtf/pdfgen/make_pdf.py')

          if not 'ana_output' in [self.pdf_rdp, self.pdf_mcp, self.pdf_oldrdp, self.pdf_oldmcp]:
            raise ApplicationConfigurationError(None,'None of the pdf inputs is set to use the make_ana.py output. Please set "pdf_rdp", "pdf_mcp", "pdf_oldrdp", or "pdf_oldmcp" to the value "ana_output"')

          for key in ['pdf_rdp', 'pdf_mcp', 'pdf_oldrdp', 'pdf_oldmcp']:
            if getattr(self, key) == 'ana_output':
              setattr(self, key, self.ana_output)
          

          argDict = { '--custom': 'pdf_custom', '--title': 'pdf_title', '--rdp': 'pdf_rdp', '--mcp': 'pdf_mcp', '--oldrdp': 'pdf_oldrdp', '--oldmcp': 'pdf_oldmcp', '--rdptitle': 'pdf_rdptitle', '--mcptitle': 'pdf_mcptitle', '--oldrdptitle': 'pdf_oldrdptitle', '--oldmcptitle': 'pdf_oldmcptitle', '--out': 'pdf_output' }
#          argDict = { '--custom': self.pdf_custom, '--title': self.pdf_title, '--rdp': self.pdf_rdp, '--mcp': self.pdf_mcp, '--oldrdp': self.pdf_oldrdp, '--oldmcp': self.pdf_oldmcp, '--rdptitle': self.pdf_rdptitle, '--mcptitle': self.pdf_mcptitle, '--oldrdptitle': self.pdf_oldrdptitle, '--oldmcptitle': self.pdf_oldmcptitle, '--out': self.pdf_output }

          for key in argDict:
            if not getattr(self, argDict[key]) == None:
              args.append(key+'='+getattr(self, argDict[key]))

          for opt in self.pdf_options:
            for key in argDict:
              if opt.find(key) > -1 and not getattr(self, argDict[key]) == None:
                raise ApplicationConfigurationError(None,'The make_pdf.py command line argument %s was set through both the ganga application variable "%s" and pdf_options "%s". Use only one of them.' % (key, argDict[key], opt))
            args.append(opt)


        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        script += ' '.join(args)+'\n'

        from Ganga.GPIDev.Lib.File import FileBuffer

        scriptname = 'make_ana.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_VFT_make_ana') #_Properties
# config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('VFT_make_ana','LSF', RTHandler)
allHandlers.add('VFT_make_ana','Local', RTHandler)
allHandlers.add('VFT_make_ana','PBS', RTHandler)
allHandlers.add('VFT_make_ana','SGE', RTHandler)
allHandlers.add('VFT_make_ana','Condor', RTHandler)
allHandlers.add('VFT_make_ana','LCG', LCGRTHandler)
allHandlers.add('VFT_make_ana','gLite', gLiteRTHandler)
allHandlers.add('VFT_make_ana','TestSubmitter', RTHandler)
allHandlers.add('VFT_make_ana','Interactive', RTHandler)
allHandlers.add('VFT_make_ana','Batch', RTHandler)
allHandlers.add('VFT_make_ana','Cronus', RTHandler)
allHandlers.add('VFT_make_ana','Remote', LCGRTHandler)
allHandlers.add('VFT_make_ana','CREAM', LCGRTHandler)




# 05 Aug 2005 - KH : Added functions getSearchPath and getScriptPath

# 16 Aug 2005 - KH : Added method RuntimePackage.loadTemplates

# 30 Aug 2006 - KH : Modified function getSearchPath, to expand ~ and
#                    environment variables, and to allow paths to be
#                    specified relative to Ganga top directory

# 19 Oct 2006 - KH : Modified function getScriptPath, to expand ~ and
#                    environment variables

# 19 Oct 2006 - KH : Generalised function getSearchPath, allowing
#                    configuration parameter defining search path
#                    to be passed as arument

from Ganga.Utility.util import importName

#from Ganga.Utility.external.ordereddict import oDict
from Ganga.Utility.external.OrderedDict import OrderedDict as oDict
allRuntimes = oDict()

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger(modulename=1)


def getScriptPath(name="", searchPath=""):
    """Determine path to a script

       Arguments:
          name       - Name of a script
          searchPath - String of colon-separated directory names,
                       defining locations to be searched for script

       If 'name' already gives the path to the script, then
       'searchPath' is ignored.

       Return value: String giving path to script (success)
                     or empty string (script not found)"""

    import os

    scriptPath = ""

    if name:
        fullName = os.path.expanduser(os.path.expandvars(name))
        if os.path.exists(fullName):
            scriptPath = fullName
        else:
            if searchPath:
                script = os.path.basename(fullName)
                pathList = searchPath.split(":")
                for directory in pathList:
                    if os.path.isdir(directory):
                        dirPath = os.path.abspath(directory)
                    if script in os.listdir(dirPath):
                        scriptPath = os.sep.join([dirPath, script])
                        break

    return scriptPath


def getSearchPath(configPar="SCRIPTS_PATH"):
    """Determine search path from configuration parameter

       Argument: 
          configPar : Name of configuration parameter defining search path

       Return value: Search path"""

    import os
    from Ganga.Utility.Config import ConfigError, getConfig

    config = getConfig("Configuration")

    utilityDir = os.path.dirname(os.path.abspath(__file__))
    gangaRoot = os.path.dirname(os.path.dirname(utilityDir))

    pathString1 = ""
    if configPar:
        try:
            pathString1 = str(config[configPar])
        except ConfigError:
            logger.error("Option '%s' not defined in 'Configuration'" %
                         configPar)

    # always have . in the path in the first position!
    pathList1 = ['.'] + pathString1.split(":")
    pathList2 = []

    for path in pathList1:
        if ("." != path):
            path = os.path.expanduser(os.path.expandvars(path))
            if (0 != path.find("/")):
                path = os.path.join(gangaRoot, path)
        pathList2.append(path)

    pathString2 = ":".join(pathList2)
    return pathString2


class RuntimePackage(object):

    def __init__(self, path):
        import os.path
        import sys
        import Ganga.Utility.Config

        self.path = os.path.normpath(path.rstrip('/'))
        self.name = os.path.basename(self.path)
        self.syspath = os.path.dirname(self.path)
        self.mod = None
        self.modpath = ''

        showpath = self.syspath
        if not showpath:
            showpath = '<defaultpath>'
        logger.debug("initializing runtime: '%s' '%s'", self.name, showpath)

        if self.name in allRuntimes:
            if allRuntimes[self.name].path != self.path:
                logger.warning('possible clash: runtime "%s" already exists at path "%s"', self.name, allRuntimes[self.name].path)

        allRuntimes[self.name] = self

        if self.syspath:
            # FIXME: not sure if I really want to modify sys.path (side effects!!)
            # allow relative paths to GANGA_PYTHONPATH
            if not os.path.isabs(self.syspath):
                self.syspath = os.path.join(
                    Ganga.Utility.Config.getConfig('System')['GANGA_PYTHONPATH'], self.syspath)
            sys.path.insert(0, self.syspath)

        # Ganga.Utility.Config.getConfig('Runtime_'+self.name)
        self.config = {}

        try:
            self.mod = __import__(self.name)
            self.modpath = os.path.dirname(
                os.path.normpath(os.path.abspath(self.mod.__file__)))
            if self.syspath:
                if self.modpath.find(self.syspath) == -1:
                    logger.warning(
                        "runtime '%s' imported from '%s' but specified path is '%s'. You might be getting different code than expected!", self.name, self.modpath, self.syspath)
            else:
                logger.debug(
                    "runtime package %s imported from %s", self.name, self.modpath)

            # import the <PACKAGE>/PACKAGE.py module
            # @see Ganga/PACKAGE.py for description of this magic module
            # in this way we enforce any initialization of module is performed
            # (e.g PackageSetup.setPlatform() is called)
            __import__(self.name + ".PACKAGE")

        except ImportError as x:
            logger.warning("cannot import runtime package %s: %s", self.name, str(x))

    def getEnvironment(self):
        # FIXME: pass the configuration object
        g = importName(self.name, 'getEnvironment')
        if g:
            return g(self.config)
        else:
            logger.debug("no environment defined for runtime package %s", self.name)
            return {}

    def loadPlugins(self):
        logger.debug("Loading Plugin: %s" % self.name)
        g = importName(self.name, 'loadPlugins')
        if g:
            g(self.config)
        else:
            logger.debug("no plugins defined for runtime package %s", self.name)
        logger.debug("Finished Plugin: %s" % self.name)

    def bootstrap(self, globals):
        try:
            import os.path

            # do not import names from BOOT file automatically, use
            # exportToGPI() function explicitly
            exec("import %s.BOOT" % self.name)
        except ImportError as x:
            logger.debug("problems with bootstrap of runtime package %s", self.name)
            logger.debug(x)
        except IOError as x:
            logger.debug("problems with bootstrap of runtime package %s", self.name)
            logger.debug(x)

    def loadNamedTemplates(self, globals, file_ext='tpl', pickle_files=False):
        try:
            import os
            from Ganga.GPIDev.Lib.Registry.RegistryUtils import establishNamedTemplates
            template_registryname = 'templates%s' % self.name.strip('Ganga')
            template_pathname = os.path.join(self.modpath, 'templates')
            if os.path.isdir(template_pathname):
                establishNamedTemplates(template_registryname,
                                        template_pathname,
                                        "Registry for '%s' NamedTemplates" % self.name.strip(
                                            'Ganga'),
                                        file_ext=file_ext,
                                        pickle_files=pickle_files)
        except:
            logger.debug('failed to load named template registry')
            raise

    def loadTemplates(self, globals):
        try:
            import os.path
            if os.path.isfile(os.path.join(self.modpath, 'TEMPLATES.py')):
                execfile(os.path.join(self.modpath, 'TEMPLATES.py'), globals)
            else:
                logger.debug("Problems adding templates for runtime package %s",
                             self.name)
        except Exception as x:
            logger.debug\
                ("Problems adding templates for runtime package %s", self.name)
            logger.debug(x)

    def shutdown(self):
        g = importName(self.name, 'shutdown')
        if g:
            g()
        else:
            from Ganga.Utility.logging import getLogger
            logger = getLogger(modulename=1)
            logger.debug("no shutdown procedure in runtime package %s", self.name)

    # this hook is called after the Ganga bootstrap procedure completed
    def postBootstrapHook(self):
        g = importName(self.name, 'postBootstrapHook')
        if g:
            g()
        else:
            logger.debug("no postBootstrapHook() in runtime package %s", self.name)




#!/usr/bin/env python
from __future__ import print_function
import shutil
import os
import time
import popen2
import glob

############################################################################################

###INLINEMODULES###
###INLINEHOSTNAMEFUNCTION###

############################################################################################

input_sandbox = ###INPUT_SANDBOX###
sharedoutputpath = ###SHAREDOUTPUTPATH###
outputpatterns = ###OUTPUTPATTERNS###
appscriptpath = ###APPSCRIPTPATH###
environment = ###ENVIRONMENT###

# jobid is a string
jobid = ###JOBID###

###PREEXECUTE###

def flush_file(f):
    f.flush()
    os.fsync(f.fileno()) #this forces a global flush (cache synchronization on AFS)

def open_file(fname):
    try:
        filehandle=open(fname,'w')
    except IOError as x:
        print('ERROR: not able to write a status file: ', fname)
        print('ERROR: ',x)
        raise
    return filehandle

statusfilename = os.path.join(sharedoutputpath,'__jobstatus__')
heartbeatfilename = os.path.join(sharedoutputpath,'__heartbeat__')

statusfile=open_file(statusfilename)
heartbeatfile=open_file(heartbeatfilename)

line='START: '+ time.strftime('%a %b %d %H:%M:%S %Y',time.gmtime(time.time())) + os.linesep
try:
    line+='PID: ' + os.getenv('###JOBIDNAME###') + os.linesep
    line+='QUEUE: ' + os.getenv('###QUEUENAME###') + os.linesep
    line+='ACTUALCE: ' + hostname() + os.linesep
except:
    pass
statusfile.writelines(line)
flush_file(statusfile)

import tarfile

# -- WARNING: get the input files including the python modules BEFORE sys.path.insert()
# -- SINCE PYTHON 2.6 THERE WAS A SUBTLE CHANGE OF SEMANTICS IN THIS AREA

for f in input_sandbox:
    getPackedInputSandbox(f)

# -- END OF MOVED CODE BLOCK

#get input files
###DOWNLOADINPUTFILES###

import sys
sys.path.insert(0, ###GANGADIR###)
sys.path.insert(0,os.path.join(os.getcwd(),PYTHON_DIR))

import subprocess

fullenvironment = os.environ.copy()
for key,value in environment.iteritems():
    fullenvironment[key] = value

sysout2 = os.dup(sys.stdout.fileno())
syserr2 = os.dup(sys.stderr.fileno())

print("--- GANGA APPLICATION OUTPUT BEGIN ---", file=sys.stdout)
print("--- GANGA APPLICATION ERROR BEGIN ---", file=sys.stdout)
flush_file(sys.stdout)
flush_file(sys.stderr)

sys.stdout=open('./__syslog__','w')
sys.stderr=sys.stdout

result = 255



try:
    child = subprocess.Popen(appscriptpath, shell=False, stdout=sysout2, stderr=syserr2, env=fullenvironment)

    while 1:
        result = child.poll()
        if result is not None:
            break
        heartbeatfile.write('.')
        flush_file(heartbeatfile)
        time.sleep(###HEARTBEATFREQUENCE###)
except Exception as x:
    print('ERROR: %s'%str(x))

flush_file(sys.stdout)
flush_file(sys.stderr)
sys.stdout=sys.__stdout__
sys.stderr=sys.__stderr__
print("--- GANGA APPLICATION OUTPUT END ---", file=sys.stdout)


try:
    filefilter
except:
    filefilter = None

from files import multi_glob, recursive_copy

createOutputSandbox(outputpatterns,filefilter,sharedoutputpath)

def printError(message):
    print(message, file=sys.stderr)

def printInfo(message):
    print(message, file=sys.stdout)

###OUTPUTUPLOADSPOSTPROCESSING###

print("--- GANGA APPLICATION ERROR END ---", file=sys.stderr)

###OUTPUTSANDBOXPOSTPROCESSING###

###POSTEXECUTE###

line='EXITCODE: ' + repr(result) + os.linesep
line+='STOP: '+time.strftime('%a %b %d %H:%M:%S %Y',time.gmtime(time.time())) + os.linesep
statusfile.writelines(line)

statusfile.close()
heartbeatfile.close()
os.unlink(heartbeatfilename)

sys.exit(result)





##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IBackend.py,v 1.2 2008-10-02 10:31:05 moscicki Exp $
##########################################################################

from Ganga.Core.exceptions import IncompleteJobSubmissionError
from Ganga.Core.GangaRepository.SubJobXMLList import SubJobXMLList
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName
from Ganga.GPIDev.Lib.Dataset import GangaDataset
from Ganga.GPIDev.Schema import Schema, Version

import Ganga.Utility.logging

from Ganga.Utility.logic import implies

import os
import itertools
import time

logger = Ganga.Utility.logging.getLogger()

class IBackend(GangaObject):

    """
    Base class for all backend objects.

    Backend  classes in  Ganga 4.0.x  have only  the  submit() method.
    This is sufficient to allow emulated bulk submission to function.

    Note  that the master_submit()  method is  always called  by the
    framework unlike the submit()  method, which is only called as
    a  part of  default  implementation of  master_submit(). If  you
    provide a special support  for bulk submission than you should
    either  call submit()  method  explicitly in  the  case of  no
    splitting, or make  the implementation of master_submit() handle
    the non-split jobs as well.

    """

    _schema = Schema(Version(0, 0), {})
    _category = 'backends'
    _hidden = 1

    # specify how the default implementation of master_prepare() method
    # creates the input sandbox
    _packed_input_sandbox = True

    def __init__(self):
        super(IBackend, self).__init__()

    def setup(self):
        """ This is a hook called for each job when Ganga.Core services are
        started up. The hook is called before the monitoring subsystem is
        enabled. This hook may be used by some backends to do specialized setup
        (e.g. to open ssh transport pipes of the Remote backend)
        """
        pass

    def _parallel_submit(self, b, sj, sc, master_input_sandbox, fqid, logger):

        try:
            sj.updateStatus('submitting')
            if b.submit(sc, master_input_sandbox):
                sj.updateStatus('submitted')
                sj.info.increment()
            else:
                raise IncompleteJobSubmissionError(fqid, 'submission failed')
        except Exception as err:
            #from Ganga.Utility.logging import log_user_exception
            sj.updateStatus('failed')

            #from Ganga.Core.exceptions import GangaException
            #if isinstance(err, GangaException):
            #    logger.error(str(err))
            #    #log_user_exception(logger, debug=True)
            #else:
            #    #log_user_exception(logger, debug=False)
            logger.error("Parallel Job Submission Failed: %s" % str(err))
        finally:
            pass

    def master_submit(self, rjobs, subjobconfigs, masterjobconfig, keep_going=False, parallel_submit=False):
        """  Submit   the  master  job  and  all   its  subjobs.   The
        masterjobconfig  is  shared,  individual  subjob  configs  are
        defined  in  subjobconfigs.   Submission  of  individual  jobs
        (not-split) also  always goes via  this method.  In  that case
        the subjobconfigs contains just one element - the job itself.

        The default  implementation of  this method emulates  the bulk
        submission  calling  a submit()  method  on individual  subjob
        objects.  If submission  of any of the subjobs  fails then the
        whole   process  is  aborted   with  IncompleteSubmissionError
        exception. The subjobs which  have already been submitted stay
        submitted.

        The default implementation does not process the masterjobconfig.
        Therefore this method may be overriden in the derived class
        in the following way:

        def master_submit(self,masterjobconfig,subjobconfigs,keep_going):
           ... 
           do_some_processsing_of(masterjobconfig)
           ...
           return IBackend.master_submit(self,subjobconfigs,masterjobconfig,keep_joing)


        Implementation note: we set keep_going to be optional in the
        signature of IBackend.master_submit() to allow the existing
        backend implementations which do not support keep_going=True
        and which at some point may call IBackend.master_submit() to
        work without change. It may sometimes be non-trivial to enable
        support for keep_going=True in some backends, even if the
        finally call IBackend.master_submit(). Therefore it is left to
        the decision of backend developer to explicitly enable the
        support for keep_going flag.

        """
        from Ganga.Core import IncompleteJobSubmissionError, GangaException
        from Ganga.Utility.logging import log_user_exception


        logger.debug("SubJobConfigs: %s" % len(subjobconfigs))
        logger.debug("rjobs: %s" % len(rjobs))
        assert(implies(rjobs, len(subjobconfigs) == len(rjobs)))

        incomplete = 0
        incomplete_subjobs = []

        def handleError(x):
            if keep_going:
                incomplete_subjobs.append(fqid)
                return False
            else:
                if incomplete:
                    raise x
                else:
                    return True

        master_input_sandbox = self.master_prepare(masterjobconfig)

        if parallel_submit:

            from Ganga.GPI import queues

            threads_before = queues.totalNumIntThreads()

            for sc, sj in zip(subjobconfigs, rjobs):

                fqid = sj.getFQID('.')
                b = sj.backend
                # FIXME would be nice to move this to the internal threads not user ones
                #from Ganga.GPIDev.Base.Proxy import stripProxy
                #all_queues = stripProxy(queues)
                #all_queues._addSystem( self._parallel_submit, ( b, sj, sc, master_input_sandbox, fqid, logger ) )
                queues._monitoring_threadpool.add_function(self._parallel_submit, (b, sj, sc, master_input_sandbox, fqid, logger))

            def subjob_status_check(rjobs):
                has_submitted = True
                for sj in rjobs:
                    if sj.status not in ["submitted","failed","completed","running","completing"]:
                        has_submitted = False
                        break
                return has_submitted

            while not subjob_status_check(rjobs):
                import time
                time.sleep(1.)

            for i in rjobs:
                if i.status in ["new", "failed"]:
                    return 0
            return 1

        for sc, sj in zip(subjobconfigs, rjobs):

            fqid = sj.getFQID('.')
            logger.info("submitting job %s to %s backend", fqid, getName(sj.backend))
            try:
                b = stripProxy(sj.backend)
                sj.updateStatus('submitting')
                if b.submit(sc, master_input_sandbox):
                    sj.updateStatus('submitted')
                    # sj._commit() # PENDING: TEMPORARY DISABLED
                    incomplete = 1
                    stripProxy(sj.info).increment()
                else:
                    if handleError(IncompleteJobSubmissionError(fqid, 'submission failed')):
                        return 0
            except Exception as x:
                #sj.updateStatus('new')
                if isType(x, GangaException):
                    logger.error(str(x))
                    log_user_exception(logger, debug=True)
                else:
                    log_user_exception(logger, debug=False)
                if handleError(IncompleteJobSubmissionError(fqid, str(x))):
                    return 0

        if incomplete_subjobs:
            raise IncompleteJobSubmissionError(
                incomplete_subjobs, 'submission failed')

        return 1

    def submit(self, subjobconfig, master_input_sandbox):
        """
        Submit an individual job. Return 1 in case of success.

        master_input_sandbox is a list of file-names which is shared by all subjobs.

        This method  is not  called directly by  the framework.  It is
        only  called by  the default  implementation  of master_submit()
        method.

        Therefore if  the implementation  of master_submit() is  able to
        cope  with submission  of  individual jobs,  then submit()  is
        redundant.

        Transition from Ganga 4.0.x:
         - subjobconfig is equvalent to jobconfig in the older interface.
         - jobid is an obsolte parameter which will be removed in the future

        """
        raise NotImplementedError

    def master_prepare(self, masterjobconfig):
        """ Prepare the master job (shared sandbox files). This method is/should be called by master_submit() exactly once.
        The input sandbox is created according to self._packed_input_sandbox flag (a class attribute)
        """
        from Ganga.GPIDev.Lib.File.OutputFileManager import getInputFilesPatterns
        from Ganga.GPIDev.Lib.File.File import File, ShareDir

        job = self.getJobObject()

        create_sandbox = job.createInputSandbox
        if self._packed_input_sandbox:
            create_sandbox = job.createPackedInputSandbox

        if masterjobconfig:
            if hasattr(job.application, 'is_prepared') and isType(job.application.is_prepared, ShareDir):
                sharedir_pred = lambda f: f.name.find(job.application.is_prepared.name) > -1
                sharedir_files = itertools.ifilter(sharedir_pred, masterjobconfig.getSandboxFiles())
                nonsharedir_files = itertools.ifilterfalse(sharedir_pred, masterjobconfig.getSandboxFiles())
            # ATLAS use bool to bypass the prepare mechanism and some ATLAS
            # apps have no is_prepared
            else:
                sharedir_files = []
                nonsharedir_files = masterjobconfig.getSandboxFiles()
            inputsandbox = create_sandbox(nonsharedir_files, master=True)
            inputsandbox.extend(
                itertools.imap(lambda f: f.name, sharedir_files))
            return inputsandbox

        tmpDir = None
        files = []
        if len(job.inputfiles) > 0 or (len(job.subjobs) == 0 and job.inputdata and\
                isType(job.inputdata, GangaDataset) and job.inputdata.treat_as_inputfiles):
            (fileNames, tmpDir) = getInputFilesPatterns(job)
            files = itertools.imap(lambda f: File(f), fileNames)
        else:
            # RTHandler is not required to produce masterjobconfig, in that
            # case just use the inputsandbox
            files = job.inputsandbox

        result = create_sandbox(files, master=True)
        if tmpDir is not None:
            import shutil
            # remove temp dir
            if os.path.exists(tmpDir):
                shutil.rmtree(tmpDir)
        return result

    def master_auto_resubmit(self, rjobs):
        """ A hook in case someone wanted to override the auto
        resubmission behaviour. Otherwise, it auto_resubmit is
        equivalent for all practical purposes to a normal resubmit (so
        it automatcially benefits from bulk resubmit if implemented).
        """
        return self.master_resubmit(rjobs)

    def check_auto_resubmit(self):
        """ A hook for the backend to check if this job can be
        automatically resubmitted.
        """
        return True

    def master_resubmit(self, rjobs, backend=None):
        """ Resubmit (previously submitted) job. Configuration phase is skipped.
        Default implementation works is an emulated-bulk operation.
        If you override this method for bulk optimization then make sure that you call updateMasterJobStatus() on the master job,
        so the master job will be monitored by the monitoring loop.
        """
        from Ganga.Core import IncompleteJobSubmissionError, GangaException
        from Ganga.Utility.logging import log_user_exception
        incomplete = 0

        def handleError(x):
            if incomplete:
                raise x
            else:
                return 0
        try:
            for sj in rjobs:
                fqid = sj.getFQID('.')
                logger.info("resubmitting job %s to %s backend", fqid, getName(sj.backend))
                try:
                    b = sj.backend
                    sj.updateStatus('submitting')
                    if backend is None:
                        result = b.resubmit()
                    else:
                        result = b.resubmit(backend=backend)
                    if result:
                        sj.updateStatus('submitted')
                        # sj._commit() # PENDING: TEMPORARY DISABLED
                        incomplete = 1
                    else:
                        return handleError(IncompleteJobSubmissionError(fqid, 'resubmission failed'))
                except Exception as x:
                    log_user_exception(
                        logger, debug=isType(x, GangaException))
                    return handleError(IncompleteJobSubmissionError(fqid, str(x)))
        finally:
            master = self.getJobObject().master
            if master:
                master.updateMasterJobStatus()
        return 1

    def resubmit(self):
        raise NotImplementedError

    def master_kill(self):
        """ Kill a job and all its subjobs. Return 1 in case of success.

        The default implementation uses the kill() method and emulates
        the bulk  operation on all subjobs.  It tries to  kill as many
        subjobs  as  possible even  if  there  are  failures.  If  the
        operation is incomplete then raise IncompleteKillError().
        """

        job = self.getJobObject()

        r = True

        if len(job.subjobs):
            problems = []
            for s in job.subjobs:
                if s.status in ['submitted', 'running']:
                    if not s.backend.kill():
                        r = False
                        problems.append(s.id)
            if not r:
                from Ganga.Core import IncompleteKillError
                raise IncompleteKillError(
                    'subjobs %s were not killed' % problems)
        else:
            r = job.backend.kill()
        return r

    def kill(self):
        """  Kill a job  (and also  all its  subjobs). This  method is
        never called by the framework directly.  It may only be called
        by the default implementation of master_kill().  """

        raise NotImplementedError

    def peek(self, filename="", command=""):
        """
        Allow viewing of job output files on the backend
        (i.e. while job is in 'running' state)

        Arguments other than self:
        filename : name of file to be viewed
                  => Path specified relative to work directory
        command  : command to be used for file viewing

        Return value: None
        """

        # This is a dummy implementation - it only provides access
        # to files in the job's output directory

        job = self.getJobObject()
        job.peek(filename=os.path.join("..", filename), command=command)
        return None

    def remove(self):
        """When the job is removed then this backend method is called.
        The primary use-case is the Remote (ssh) backend. """
        pass

    def getStateTime(self, status):
        """Get the timestamps for the job's transitions into the 'running' and 'completed' states.
        """
        return None

    def timedetails(self):
        """Returns all available backend specific timestamps.
        """
        pass

    @staticmethod
    def master_updateMonitoringInformation(jobs):
        """ Update monitoring information for  jobs: jobs is a list of
        jobs  in   this  backend  which   require  monitoring  (either
        'submitted' or 'running' state).  The jobs list never contains
        the subjobs.

        The default implementation  iterates  over subjobs and calls
        updateMonitoringInformation().
        """

        ## Have to import here so it's actually defined
        from Ganga.Core import monitoring_component

        logger.debug("Running Monitoring for Jobs: %s" % str([j.getFQID('.') for j in jobs]))

        ## Only process 10 files from the backend at once
        #blocks_of_size = 10
        try:
            from Ganga.Utility.Config import getConfig
            blocks_of_size = getConfig('PollThread')['numParallelJobs']
        except Exception as err:
            logger.debug("Problem with PollThread Config, defaulting to block size of 5 in master_updateMon...")
            logger.debug("Error: %s" % str(err))
            blocks_of_size = 5
        ## Separate different backends implicitly
        simple_jobs = {}

        # FIXME Add some check for (sub)jobs which are in a transient state but
        # are not locked by an active session of ganga

        for j in jobs:
            ## All subjobs should have same backend
            if len(j.subjobs) > 0:
                #logger.info("Looking for sj")
                monitorable_subjob_ids = []

                if isType(j.subjobs, SubJobXMLList):
                    cache = j.subjobs.getAllCachedData()
                    for sj_id in range(0,len(j.subjobs)):
                        if cache[sj_id]['status'] in ['submitted', 'running']:
                            if j.subjobs.isLoaded(sj_id):
                                ## SJ may have changed from cache in memory
                                this_sj = j.subjobs(sj_id)
                                if this_sj.status in ['submitted', 'running']:
                                    monitorable_subjob_ids.append(sj_id)
                            else:
                                monitorable_subjob_ids.append(sj_id)
                else:
                    for sj in j.subjobs:
                        if sj.status in ['submitted', 'running']:
                            monitorable_subjob_ids.append(sj.id)

                #logger.info('Monitoring subjobs: %s', str(monitorable_subjob_ids)

                if not monitorable_subjob_ids:
                    continue

                stripProxy(j)._getWriteAccess()

                #logger.info("Dividing")

                monitorable_blocks = []
                temp_block = []

                for this_sj_id in monitorable_subjob_ids:
                    temp_block.append(this_sj_id)
                    if len(temp_block) == blocks_of_size:
                        monitorable_blocks.append(temp_block)
                        temp_block = []

                if temp_block:
                    monitorable_blocks.append(temp_block)
                    temp_block = []

                for this_block in monitorable_blocks:

                    if monitoring_component and not monitoring_component.isEnabled(False) or not monitoring_component:
                        break

                    try:
                        subjobs_to_monitor = []
                        for sj_id in this_block:
                            subjobs_to_monitor.append(j.subjobs[sj_id])
                        stripProxy(j.backend).updateMonitoringInformation(subjobs_to_monitor)
                    except Exception as err:
                        logger.error("Monitoring Error: %s" % str(err))
                    j.updateMasterJobStatus()

                ## NB ONLY THE MASTER JOB IS KNOWN TO THE JOB REPO!!!
                stripProxy(j)._setDirty()
            else:
                backend_name = getName(j.backend)
                if backend_name not in simple_jobs.keys():
                    simple_jobs[backend_name] = []
                simple_jobs[backend_name].append(j)

        if len(simple_jobs.keys()) > 0:
            for this_backend in simple_jobs.keys():
                logger.debug('Monitoring jobs: %s', repr([jj._repr() for jj in simple_jobs[this_backend]]))

                for this_job in simple_jobs[this_backend]:
                    stripProxy(this_job)._getWriteAccess()
                stripProxy(simple_jobs[this_backend][0].backend).updateMonitoringInformation(simple_jobs[this_backend])

                for this_job in simple_jobs[this_backend]:
                    stripProxy(this_job)._setDirty()

        logger.debug("Finished Monitoring request")

    @staticmethod
    def updateMonitoringInformation(jobs):
        """ Update monitoring information for individual jobs: jobs is
        a  list which  may contain  subjobs as  well as  the non-split
        jobs.  This method is  never called by the framework directly.
        It  may  only  be  called  by the  default  implementation  of
        master_updateMonitoringInformation().
        """

        raise NotImplementedError



"""Core submitter IAction implementation.

The CoreSubmitter class provides a generic submitter implementation.
 
"""

from GangaRobot.Lib.Base.BaseSubmitter import BaseSubmitter
from Ganga.Utility.logging import getLogger
from Ganga.Utility import files
from Ganga.GPI import *
import os

logger = getLogger()


class CoreSubmitter(BaseSubmitter):

    """Core submitter IAction implementation.
    
    A submitter which submits exported jobs. See handlesubmit() for details.
    
    This submitter can be reused by adding it to the chain of extractors
    initialised in the constructor of any implementation of BaseExtractor.
    e.g.
    def __init__(self):
        self.chain = [CoreSubmitter(), self]
    
    """

    def handlesubmit(self, jobids, runid):
        """Submits exported jobs as identified by a list of path patterns.
        
        Keyword arguments:
        jobids -- A list of the submitted job ids.
        runid -- A UTC ID string which identifies the run.
        
        The following parameters are loaded from the Robot configuration:
        CoreSubmitter_Patterns e.g. ['GangaRobot/Lib/Core/jobs/core_jobs.txt']

        All files matched by the patterns are assumed to contain single or
        multiple exported jobs, as created using the GPI function 'export'.

        Relative paths are evaluated against the current working directory and
        the Ganga python root, with duplicates being removed.
        
        The jobs are loaded from the files into the job repository, submitted
        and their ids are added to the jobids list.
        
        """
        # get configuration properties
        patterns = self.getoption('CoreSubmitter_Patterns')
        logger.info("Searching for job files matching patterns %s.", patterns)
        matches = self._getmatches(patterns)
        logger.info("Found %d matching job files.", len(matches))

        for match in matches:
            jobs = load(match)
            logger.info("Loaded %d jobs from '%s'.", len(jobs), match)
            for j in jobs:
                j.submit()
                jobids.append(j.id)
        
        
    def _getmatches(self, patterns):
        """Return a list of paths matching the list of path patterns.
        
        Keyword arguments:
        patterns -- A list of path patterns, identifying exported job files.

        Relative paths are evaluated against the current working directory and
        against the Ganga python root, i.e. ganga/python/. Duplicate matches are
        removed (based on absolute file name).
        
        """
        # expand patterns
        patterns = [files.expandfilename(p) for p in patterns]
        # add local patterns absolute or relative to cwd
        cwd = os.getcwd()
        localpatterns = [os.path.join(cwd, p) for p in patterns]
        # add local patterns absolute or relative to Ganga python root
        gangapythonroot = config['System']['GANGA_PYTHONPATH']
        syspatterns = [os.path.join(gangapythonroot, p) for p in patterns]
        # find all matches removing duplicates
        matches = files.multi_glob(localpatterns + syspatterns)
        return matches



from Ganga.GPIDev.Base.Proxy import addProxy
from Ganga.GPIDev.Schema import Schema, Version
from Ganga.GPIDev.Lib.Tasks.IUnit import IUnit
import copy
import Ganga.GPI as GPI


class CoreUnit(IUnit):
    _schema = Schema(Version(1, 0), dict(IUnit._schema.datadict.items() + {
    }.items()))

    _category = 'units'
    _name = 'CoreUnit'
    _exportmethods = IUnit._exportmethods + []

    def __init__(self):
        super(CoreUnit, self).__init__()

    def createNewJob(self):
        """Create any jobs required for this unit"""
        j = GPI.Job()

        j.backend = self._getParent().backend.clone()

        # copy form ourselves or the parent transform depending on what's
        # specified
        fields = ['application', 'splitter', 'inputfiles',
                  'inputdata', 'inputsandbox', 'outputfiles', 'postprocessors']

        for f in fields:

            if (f == "postprocessors" and len(getattr(self, f).process_objects) > 0):
                j.postprocessors = copy.deepcopy(addProxy(self).postprocessors)
            elif (f != "postprocessors" and getattr(self, f)):
                setattr(j, f, copy.deepcopy(getattr(self, f)))
            elif (f == "postprocessors" and len(getattr(self._getParent(), f).process_objects) > 0):
                j.postprocessors = copy.deepcopy(
                    addProxy(self._getParent()).postprocessors)
            elif (f != "postprocessors" and getattr(self._getParent(), f)):
                setattr(j, f, copy.deepcopy(getattr(self._getParent(), f)))

        return j



#! /usr/bin/env python

# -------------------------------------------------------------
# globals and imports
import os, sys, pickle
from commands import getstatusoutput
import dq2
import shutil
from dq2.clientapi.DQ2 import DQ2
from struct import *

_refCache = {}
_streamRef = ''
_datasetType = 'AOD'
_dq = DQ2 ()

def findReferences( infile ):
    """Find the references from this input TAG file"""
    global _refCache, _streamRef, _datasetType

    # Find the links to the stream required
    cmd = "CollListFileGUID -src " + infile + " RootCollection -queryopt "+ _streamRef +" | cut -d' ' -f 1"
    rc, out = getstatusoutput(cmd)
    if (rc!=0) or (out.find("command not found")!=-1):
        print "ERROR: error during CollListFileGUID:"
        print out
        return {}

    # get a list of the guids
    ref_guids = []
    refs = []
    for ln in out.split('\n'):
        try:
            ref_guids.append(ln.split()[0])
        except:
            continue

    #print "GUIDs referenced by " + infile
    #print repr(ref_guids)
    if len(ref_guids) == 0:
        print "ERROR: Could not find any references of type '%s' in file %s" % (_streamRef, infile)
        return {}

    # go over the guids and find all references
    for ref_guid in ref_guids:
        #print "Getting dataset referenced by GUID %s" % ref_guid
        
        ref_name = ''
        ref_dataset = ''

        # check the cache first
        for name in _refCache:
            if ref_guid in _refCache[name]:

                ref_name = _refCache[name][ref_guid]['lfn']
                ref_dataset = name
                refs.append( [_refCache[name][ref_guid]['lfn'], name, ref_guid ] )
                    
                
        if ref_name != '' and ref_dataset != '':
            continue

        # not in cache so try to find a valid dataset
        ref_vuids = _dq.contentClient.queryDatasetsWithFileByGUID(ref_guid)
        if len(ref_vuids) == 0:
            continue

        for ref_vuid in ref_vuids:
                    
            try:
                #print "Trying VUID %s for GUID %s... " % (ref_vuid, ref_guid)
                ref_dataset = _dq.repositoryClient.resolveVUID(ref_vuid)
                ref_name = ref_dataset.get('dsn')

                if ref_name != '' and ref_name[:8].find("panda.") != 0 and len(_dq.listDatasetReplicas(ref_name)) != 0 and ((ref_name.find("." + _datasetType + ".") != -1) or ((_datasetType == 'RAW') and ref_name.find("." + _datasetType) != -1)):
                    break
                else:
                    ref_name = ''
                    
            except dq2.repository.DQRepositoryException.DQUnknownDatasetException:
                pass
                #print "ERROR Finding dataset for vuid for " + ref_vuid

        if ref_name == '':
            continue

        # store useful stuff
        ref_files = _dq.listFilesInDataset(ref_name)
        if ref_guid in ref_files[0]:
            refs.append([ ref_files[0][ref_guid]['lfn'], ref_name, ref_guid ] )
            #print "Found dataset ref: %s, %s" % (ref_name, ref_files[0][ref_guid]['lfn'])

            # cache this dataset in case it's referenced elsewhere...
            _refCache[ref_name] = ref_files[0]


    return refs

def findReferences2( infile ):
    """Find the references from this input TAG file"""
    global _refCache, _streamRef, _datasetType

    # Find the links to the stream required
    cmd = "CollListToken -src " + infile + " RootCollection | grep -E \"Tokens|StreamTAG|%s\"" % _streamRef
    rc, out = getstatusoutput(cmd)

    if (rc!=0):
        print "ERROR: error during CollListFileGUID:"
        print out
        return {}

    # get a list of the guids
    ref_guids = {}
    refs = []
    all_ref_guids = []
    guid = ""
    tag_guid = ""
    for ln in out.split('\n'):

        # go over and find each event
        if ln.find("Tokens:") != -1:

            if tag_guid != "" and guid != "":
                if not tag_guid in ref_guids.keys():
                    ref_guids[tag_guid] = []

                if not guid in ref_guids[tag_guid]:
                    ref_guids[tag_guid].append(guid)

            guid = ""
            tag_guid = ""
            continue

        if ln.find(_streamRef) != -1 and ln.find("INFO") == -1:
            guid = ln[ ln.find("DB=")+3 : ln.find("]") ]

        if ln.find("StreamTAG") != -1 and ln.find("INFO") == -1:
            tag_guid = ln[ ln.find("DB=")+3 : ln.find("]") ]

    #print "GUIDs referenced by " + infile
    #print repr(ref_guids)
    if len(ref_guids) == 0:
        print "ERROR: Could not find any references of type '%s' in file %s" % (_streamRef, infile)
        return {}


    # go over the guids and find all references -  TAG files first
    tag_info = {}
    for ref_guid in ref_guids:
        #print "Getting dataset referenced by GUID %s" % ref_guid
        
        ref_name = ''
        ref_dataset = ''
        
        # check the cache first
        for name in _refCache:
            if ref_guid in _refCache[name]:

                ref_name = _refCache[name][ref_guid]['lfn']
                ref_dataset = name
                
                #refs.append( [_refCache[name][ref_guid]['lfn'], name, ref_guid ] )
                tag_info[_refCache[name][ref_guid]['lfn']] = {}
                tag_info[_refCache[name][ref_guid]['lfn']]['dataset'] = name
                tag_info[_refCache[name][ref_guid]['lfn']]['guid'] = ref_guid
                tag_info[ref_files[0][ref_guid]['lfn']]['refs'] = {}
                
        if ref_name != '' and ref_dataset != '':
            continue

        # not in cache so try to find a valid dataset
        ref_vuids = _dq.contentClient.queryDatasetsWithFileByGUID(ref_guid)
        #print ref_vuids
        if len(ref_vuids) == 0:
            continue

        for ref_vuid in ref_vuids:
                    
            try:
                #print "Trying VUID %s for GUID %s... " % (ref_vuid, ref_guid)
                ref_dataset = _dq.repositoryClient.resolveVUID(ref_vuid)
                ref_name = ref_dataset.get('dsn')
                if ref_name != '' and len(_dq.listDatasetReplicas(ref_name)) != 0 and ref_name.find(".TAG.") != -1:
                    break
                else:
                    ref_name = ''
                    
            except dq2.repository.DQRepositoryException.DQUnknownDatasetException:
                print "ERROR Finding dataset for vuid for " + ref_vuid
            
        if ref_name == '':
            continue
        
        # store useful stuff
        ref_files = _dq.listFilesInDataset(ref_name)
        if ref_guid in ref_files[0]:
            #refs.append([ ref_files[0][ref_guid]['lfn'], ref_name, ref_guid ] )
            tag_info[ref_files[0][ref_guid]['lfn']] = {}
            tag_info[ref_files[0][ref_guid]['lfn']]['dataset'] = ref_name
            tag_info[ref_files[0][ref_guid]['lfn']]['guid'] = ref_guid
            tag_info[ref_files[0][ref_guid]['lfn']]['refs'] = {}
            
            print "Found dataset ref: %s, %s" % (ref_name, ref_files[0][ref_guid]['lfn'])

            # cache this dataset in case it's referenced elsewhere...
            _refCache[ref_name] = ref_files[0]


    # NOW do the data files
    for tag_name in tag_info:
        
        tag_guid = tag_info[tag_name]['guid']
        
        for ref_guid in ref_guids[tag_guid]:

            print "Getting dataset referenced by GUID %s" % ref_guid

            ref_name = ''
            ref_dataset = ''

            # check the cache first
            for name in _refCache:
                if ref_guid in _refCache[name]:

                    ref_name = _refCache[name][ref_guid]['lfn']
                    ref_dataset = name

                    tag_info[tag_name]['refs'].append( [_refCache[name][ref_guid]['lfn'], name, ref_guid ] )


            if ref_name != '' and ref_dataset != '':
                continue

            # not in cache so try to find a valid dataset
            ref_vuids = _dq.contentClient.queryDatasetsWithFileByGUID(ref_guid)
            if len(ref_vuids) == 0:
                continue

            for ref_vuid in ref_vuids:

                try:
                    print "Trying VUID %s for GUID %s... " % (ref_vuid, ref_guid)
                    ref_dataset = _dq.repositoryClient.resolveVUID(ref_vuid)
                    ref_name = ref_dataset.get('dsn')
                    if ref_name != '' and len(_dq.listDatasetReplicas(ref_name)) != 0 and ref_name.find(".TAG.") != -1:
                        break
                    else:
                        ref_name = ''

                except dq2.repository.DQRepositoryException.DQUnknownDatasetException:
                    print "ERROR Finding dataset for vuid for " + ref_vuid

            if ref_name == '':
                continue

            # store useful stuff
            ref_files = _dq.listFilesInDataset(ref_name)
            if ref_guid in ref_files[0]:
                tag_info[tag_name]['refs'].append( [_refCache[name][ref_guid]['lfn'], name, ref_guid ] )

                print "Found dataset ref: %s, %s" % (ref_name, ref_files[0][ref_guid]['lfn'])

                # cache this dataset in case it's referenced elsewhere...
                _refCache[ref_name] = ref_files[0]


    
    print tag_info
    
    return {}

#-------------------------------------------
# Go through the input file list and get TAG info
# stream_ref: AOD/ESD/RAW
def createTagInfo( stream_ref, infiles ):

    global _refCache, _streamRef, _datasetType
    
    if not stream_ref in ['AOD','ESD','RAW']:
        raise TypeError("Incorrect stream ref specification %s" % stream_ref)

    _datasetType = stream_ref
    _streamRef = 'Stream'+_datasetType+'_ref'

    taginfo = {}
    taglfns = infiles
    tag_ref_files = []

    #-------------------------------------------
    # sort out the files for local running
    num = 0
    file_refs = {}
    print_num = 0
    
    for f in taglfns:        
        print f
        if os.access( f, os.R_OK ):

            if not os.path.exists( os.path.basename(f) + ".root"):
                os.symlink(f, os.path.basename(f) + ".root")

            # first, list all the guids that we care about
            refs = findReferences(os.path.basename(f))

            # try to match on names
            tagCache = {}
            for ref in refs:
                file_done = False
                tagds = ref[1].replace(_datasetType, 'TAG').replace('.recon.', '.merge.')
                poss_tagds =_dq.listDatasets(tagds + "*")

                del_list = []
                for tagds2 in poss_tagds.keys():
                    if tagds2.find("/") != -1 or (tagds.find("_tid") == -1 and tagds2.find("_tid") != -1) or tagds2.find("_sub") != -1 or tagds2.find("TAG.x") != -1:
                        del_list.append(tagds2)

                for tagds2 in del_list:
                    del poss_tagds[tagds2]

                if len(poss_tagds.keys()) == 0:
                    print "ERROR: Couldn't find matching TAG dataset."
                    return {}
                elif len(poss_tagds.keys()) > 1:
                    if print_num < 10:
                        print "WARNING: More than one TAG dataset found. Picking the first"
                        print "Possible Datasets were %s" % poss_tagds.keys()
                        print_num += 1
                        if print_num == 10:
                            print "WARNING: Reached printout limit..."
                            
                tagds = poss_tagds.keys()[0]
                
                if not tagds in tagCache:                    
                    tagCache[tagds] = _dq.listFilesInDataset(tagds)[0]


                for tf in tagCache[tagds]:
                    if ref[0].replace(_datasetType, 'TAG') == tagCache[tagds][tf]['lfn']:
                        taginfo[ref[0].replace(_datasetType, 'TAG')] = {}
                        taginfo[ref[0].replace(_datasetType, 'TAG')]['dataset'] = tagds
                        taginfo[ref[0].replace(_datasetType, 'TAG')]['guid'] = tf
                        if not 'refs' in taginfo[ref[0].replace(_datasetType, 'TAG')].keys():
                            taginfo[ref[0].replace(_datasetType, 'TAG')]['refs'] = []

                        taginfo[ref[0].replace(_datasetType, 'TAG')]['refs'].append(ref)
                        file_done = True

                if not file_done:
                    if print_num < 10:
                        print "WARNING: No direct matching of files from data to TAG."
                        print_num += 1
                        if print_num == 10:
                            print "WARNING: Reached printout limit..."
                    
                    for tf in tagCache[tagds]:
                        if not tagCache[tagds][tf]['lfn'] in taginfo.keys():
                            taginfo[ tagCache[tagds][tf]['lfn'] ] = {}
                        taginfo[tagCache[tagds][tf]['lfn']]['dataset'] = tagds
                        taginfo[tagCache[tagds][tf]['lfn']]['guid'] = tf
                        if not 'refs' in taginfo[tagCache[tagds][tf]['lfn']].keys():
                            taginfo[tagCache[tagds][tf]['lfn']]['refs'] = []

                        taginfo[tagCache[tagds][tf]['lfn']]['refs'].append(ref)

                    #raise TypeError("COULDN'T FIND LINK FOR FILE:  %s (%s)" % (ref[0], ref[1]))
                    #tag_info = {}
                    #return {}

    return taginfo

if __name__ == "__main__":

    taglfns = [ line.strip() for line in file('input_files') ]

    if 'STREAM_REF' in os.environ:
        taginfo = createTagInfo( os.environ['STREAM_REF'], taglfns )
    else:
        taginfo = createTagInfo( 'AOD', taglfns )

    if len(taginfo) == 0:
        print "ERROR: Empty taginfo structure"
        pickle.dump( taginfo, open("taginfo.pkl", "w") )
        sys.exit(-1)
    
    pickle.dump( taginfo, open("taginfo.pkl", "w") )
        



from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks.IUnit import IUnit
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base.Proxy import addProxy, stripProxy

import os
from copy import deepcopy

class ND280Unit(IUnit):
   _schema = Schema(Version(1,0), dict(IUnit._schema.datadict.items() + {
    }.items()))

   _category = 'units'
   _name = 'ND280Unit'
   _exportmethods = IUnit._exportmethods + [ ]

   def __init__(self):
      super(ND280Unit, self).__init__()
      
   def createNewJob(self):
      """Create any jobs required for this unit"""      
      j = GPI.Job()
      j._impl.backend = self._getParent().backend.clone()
      j._impl.application = self._getParent().application.clone()
      if not self.inputdata == None:
        j.inputdata = self.inputdata.clone()

      trf = self._getParent()
      task = trf._getParent()

      # copy across the outputfiles
      for f in trf.outputfiles:
         j.outputfiles += [f.clone()]

      j.inputsandbox = trf.inputsandbox

      # Sort out the splitter
      if trf.splitter:
         j.splitter = trf.splitter.clone()

      # Postprocessors
      for pp in trf.postprocessors:
         j.postprocessors.append(deepcopy(pp))
         
      return j

   def checkMajorResubmit(self, job):
      """Check if a failed job shold be 'rebrokered' (new job created) rather than just resubmitted"""
      return False

   def majorResubmit(self, job):
      """Do any bookkeeping before a major resubmit is done"""
      super(ND280Unit,self).majorResubmit(job)

   def reset(self):
      """Reset the unit completely"""
      super(ND280Unit,self).reset()

   def updateStatus(self, status):
      """Update status hook"""
      super(ND280Unit,self).updateStatus(status)

   def checkForSubmission(self):
      """Additional checks for unit submission"""

      # call the base class
      if not super(ND280Unit,self).checkForSubmission():
         return False

      return True

   def copyOutput(self):
      """Copy the output data to local storage"""

      job = GPI.jobs(self.active_job_ids[0])
      
      if self.copy_output._name != "TaskLocalCopy" or job.outputdata._impl._name != "DQ2OutputDataset":
         logger.error("Cannot transfer from DS type '%s' to '%s'. Please contact plugin developer." % (job.outputdata._name, self.copy_output._name))
         return False

      # check which fies still need downloading
      to_download = []
      for f in job.outputfiles:
         
         # check for REs
         if self.copy_output.isValid( os.path.join(f.localDir, f.namePattern)) and not self.copy_output.isDownloaded( os.path.join(f.localDir, f.namePattern)):
            to_download.append(f)

      # is everything downloaded?
      if len(to_download) == 0:
         return True

      # nope, so pick the requested number and off we go
      for f in to_download:
         f.get()

      return False



from __future__ import absolute_import
# Import classes that should be in the Tasks namespace
# Import the list of tasks and the task and abstract job definition
#from TaskList import TaskList

from Ganga.Core.GangaRepository import addRegistry
from .TaskRegistry import TaskRegistry

myTaskRegistry = TaskRegistry("tasks", "Tasks Registry")

addRegistry(myTaskRegistry)

def stopTasks():
    global myTaskRegistry
    myTaskRegistry.stop()

# Tasks
from .Task import Task
from .Transform import Transform

from .ITask import ITask
from .ITransform import ITransform
from .TaskChainInput import TaskChainInput
from .TaskLocalCopy import TaskLocalCopy

from .CoreTask import CoreTask
from .CoreTransform import CoreTransform
from .CoreUnit import CoreUnit

# Start Logger
#import Ganga.Utility.logging
#logger = Ganga.Utility.logging.getLogger()
from .common import logger

from .TaskApplication import ExecutableTask, ArgSplitterTask





#!/usr/bin/env python

######################################################
# sillyPythonWrapper.py
# ---------
# Author: Matt Mottram
#         <m.mottram@sussex.ac.uk>
#
# Description:
#    Sets up the correct python for SNO+ LCG jobs.
#
# Not really a wrapper, but required for ratProdRunner and
# ratRunner jobs that are submitted to the LCG.  RAT needs
# python 2.6 or later (installed in the VO_SNOPLUS_SNOLAB_CA_SW_DIR)
# but LCG backends default to python 2.4.3. 
#
# Importantly: functions in this script must work on older (e.g. 2.4)
# versions of python.
#
# Revision History:
#  - 03/11/12: M. Mottram: first revision with proper documentation!
#  - 12/12/12: M. Mottram: added wrapper for westgrid.
#  - 26/08/14: M. Mottram: updated for use with CVMFS; pep8
# 
######################################################

import os
import sys
import subprocess
import optparse
import job_tools

def check_el_version():
    '''Test the Enterprise Linux version.'''
    import platform
    # Could also check /etc/system-release
    bits = platform.release().split('.')
    release = None
    for b in bits:
        if "el" in b:
            try:
                release = int(b.strip('el'))
            except (ValueError, TypeError):
                print "Unknown system: " + platform.release()
                raise
    if release==5 or release==6:
        return release
    # Arrived here: do not know the system
    raise Exception("Unknown system: "+platform.release())


def run_script_lcg(script_name, args):
    '''Setup environment and run script on LCG worker node.'''
    version = check_el_version()
    # Only needed if CVMFS software is being used
    script = 'source $VO_SNOPLUS_SNOLAB_CA_SW_DIR/sl' + str(version) + '/env_cvmfs.sh\n'
    # hack on next line; uncomment for sites where VO_SNOPLUS_SNOLAB_CA_SW_DIR does not point to cvmfs
    # script = 'source /cvmfs/snoplus.gridpp.ac.uk/sl' + str(version) + '/env_cvmfs.sh\n'
    script += 'python %s ' % script_name
    script += ' '.join(str(a) for a in args)
    # As long as the python command is the final one, the exception raising should work!
    job_tools.execute_complex(script, os.getcwd(), True)


def run_script_wg(script_name, args):
    '''Setup environment and run script on WG worker node.'''
    script = 'module load application/python/2.7.3 \n'
    # Unsure if any of these are still required
    script += 'source /opt/exp_software/atlas.glite/setup_WN.sh \n'
    script += 'export LFC_HOST=lfc.gridpp.rl.ac.uk \n'
    script += 'export MYPROXY_SERVER=myproxy.westgrid.ca \n'
    script += 'python %s ' % script_name
    script += ' '.join(str(a) for a in args)
    # As long as the python command is the final one, the exception raising should work!
    job_tools.execute_complex(script, os.getcwd(), True)


def run_script(script_name, args, env_file):
    '''Setup environment and run script for an arbitrary backend.

    Knowledge of environment is required.'''
    script = file(env_file, 'r').read()
    script += '\n' # Just in case no eof
    script += 'python %s ' % script_name
    script += ' '.join(str(a) for a in args)
    # As long as the python command is the final one, the exception raising should work!
    job_tools.execute_complex(script, os.getcwd(), True, True)


if __name__=='__main__':
    parser = optparse.OptionParser("usage: <options> script location")
    parser.add_option('-s', type='string', dest='script', help='python script to run')
    parser.add_option('-a', type='string', default='', dest='args', help='python script args')
    parser.add_option('-l', type='string', dest='location', help='where am i running (lcg/wg)')
    parser.add_option('-f', type='string', dest='env_file', help='send a file with the appropriate environment')
    (options, args) = parser.parse_args()

    try:
        script = args[0]
        location = args[1]
    except IndexError:
        parser.print_help()
        raise

    script_args = options.args.split() # passed a string (e.g. "-s foo -t bar")
    if location=='lcg':
        run_script_lcg(script, script_args)
    elif location=='wg':
        run_script_wg(script, script_args)
    elif location=='misc':
        run_script(script, options.env_file, script_args)
    else:
        parser.print_help()
        print 'unknown lcoation:', location
        sys.exit(1)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
from Ganga.GPIDev.Base import GangaObject
from LHCbDatasetUtils import strToDataFile

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class OutputData(GangaObject):

    '''Class for handling outputdata for LHCb jobs.

    Example Usage:
    od = OutputData(["file.1","file.2"])
    od[0] # "file.1"
    [...etc...]
    '''
    schema = {}
    schema['files'] = SimpleItem(defvalue=[], typelist=['str'], sequence=1)
    schema['location'] = SimpleItem(defvalue='', typelist=['str'])
    _schema = Schema(Version(1, 1), schema)
    _category = 'datasets'
    _name = "OutputData"
    _exportmethods = ['__len__', '__getitem__']

    def __init__(self, files=None):
        if files is None:
            files = []
        super(OutputData, self).__init__()
        self.files = files

    def __construct__(self, args):
        if (len(args) != 1) or (type(args[0]) not in [list, tuple]):
            super(OutputData, self).__construct__(args)
        else:
            self.files = args[0]

    def __len__(self):
        """The number of files in the dataset."""
        result = 0
        if self.files:
            result = len(self.files)
        return result

    def __nonzero__(self):
        """This is always True, as with an object."""
        return True

    def __getitem__(self, i):
        '''Proivdes scripting (e.g. od[2] returns the 3rd file name) '''
        if type(i) == type(slice(0)):
            return GPIProxyObjectFactory(OutputData(files=self.files[i]))
        else:
            return self.files[i]

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: TagPrepare.py,v 1.1 2009-21-09 13:40:03 mslater Exp $
###############################################################################
# Tag Prepare Application

import os, re, commands, string

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config import makeConfig, ConfigError
from Ganga.Utility.logging import getLogger
from Ganga.Utility.files import expandfilename

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaAtlas.Lib.ATLASDataset import filecheck

from Ganga.Lib.Mergers.Merger import *

__directory__ = os.path.dirname(__file__)
logger = getLogger()

class TagPrepare(IApplication):
    """Run a preparation job for TAG analysis"""

    _schema = Schema(Version(2,2), {
        'atlas_release'          : SimpleItem(defvalue='',doc='ATLAS Software Release'),
        'tag_info'               : SimpleItem(defvalue={},doc='Returned TAG info'),
        'stream_ref'             : SimpleItem(defvalue='AOD',doc='The stream reference to use (AOD, ESD).'),
        'max_num_refs'           : SimpleItem(defvalue=3,doc='The maximum number of references retrieve per TAG file (defaults to 3).'),
        'atlas_cmtconfig'        : SimpleItem(defvalue='',doc='ATLAS CMTCONFIG environment variable'),
        'lcg_prepare'            : SimpleItem(defvalue=False,doc='Prepare an LCG TAG submission rather than Panda')
              })
                     
    _category = 'applications'
    _name = 'TagPrepare'
    _exportmethods = ['prepare_user_area']

    def postprocess(self):
        """Load in and sort out the tag info"""
        from Ganga.GPIDev.Lib.Job import Job
        job = self.getJobObject()

        if job.subjobs:
            # collect the subjob info
            self.tag_info = {}
            for j in job.subjobs:
                self.tag_info.update( j.application.tag_info )

            # loop over them and check for _subxxxx style datasets
            ref_datasets = []
            for tf in self.tag_info:
                if not self.tag_info[tf]['refs'][0][1] in ref_datasets:
                    ref_datasets.append( self.tag_info[tf]['refs'][0][1] )

            for tf in self.tag_info:
                dataset = self.tag_info[tf]['refs'][0][1]
                for ref in ref_datasets:
                    if dataset.find( (ref + '_sub') ) != -1:
                        self.tag_info[tf]['refs'][0][1] = ref
        else:
            import pickle
            self.tag_info = pickle.load( open(os.path.join( job.outputdir, "taginfo.pkl")) )
            if os.path.exists(os.path.join( job.outputdir, 'subcoll.tar.gz')):
                os.system('cd %s && tar -zxf %s/subcoll.tar.gz' % (job.outputdir, job.outputdir) )
            
    def configure(self,masterappconfig):
        logger.debug('TagPrepare configure called')
        return (None,None)

    def master_configure(self):

        logger.debug('TagPrepare master_configure called')
        job = self.getJobObject()

        if job.inputdata:
            if job.inputdata._name == 'DQ2Dataset':
                if job.inputdata.dataset and not job.inputdata.dataset_exists():
                    raise ApplicationConfigurationError(None,'DQ2 input dataset %s does not exist.' % job.inputdata.dataset)
            elif job.inputdata._name == 'ATLASLocalDataset':
                if len(job.inputdata.names) == 0:
                    raise ApplicationConfigurationError(None,'No inputdata specified in ATLASLocalDataset')

        if self.atlas_release == '':
            raise ApplicationConfigurationError(None,'No Atlas release specefied.')
        
        return (0,None)

    def prepare_user_area(self):
        "Copy appropriate files to a user area and return the required list"
        import shutil
        
        # copy the overall TAG info
        job = self.getJobObject()
        files_to_copy = ['%s/subcoll.tar.gz' % job.outputdir]

        # Now the reference files
        ref_files = []
        for tag in self.tag_info:
            toks = tag.split('.')
            stripname = '.'.join( toks[: len(toks) - 3] )

            if not stripname in ref_files:
                ref_files.append( stripname )

        for f in ref_files:
            files_to_copy.append( '%s/%s.ref.root' % (job.outputdir, f) )

        filelist = []
        logger.warning('Copying the files required for TAG running to the current directory...')
        for f in files_to_copy:
            shutil.copyfile( f, '%s/%s' % (os.getcwd(), os.path.basename(f)))
            filelist.append( os.path.basename(f) )

        return filelist
    
from Ganga.GPIDev.Adapters.IMerger import IMerger
from commands import getstatusoutput    
import threading
from GangaAtlas.Lib.ATLASDataset import Download
from GangaAtlas.Lib.ATLASDataset import filecheck



from GangaGaudi.Lib.Splitters.GaudiInputDataSplitter import GaudiInputDataSplitter
from GangaDirac.Lib.Splitters.SplitterUtils import DiracSplitter
from GangaDirac.Lib.Files.DiracFile import DiracFile
from Ganga.GPIDev.Adapters.ISplitter import SplittingError
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
from Ganga.Utility.Config import getConfig
from Ganga.Utility.files import expandfilename
from Ganga.GPIDev.Base.Proxy import stripProxy
import Ganga.Utility.logging
from Ganga.GPIDev.Lib.Job import Job
logger = Ganga.Utility.logging.getLogger()
import os
import copy
import pickle


class SplitFilesBySize(GaudiInputDataSplitter):

    """Splits a job into sub-jobs by partitioning the input data

    SplitFilesBySize can be used to split a job into multiple subjobs, where
    each subjob gets an unique subset of the inputdata files.
    """
    _name = 'SplitFilesBySize'
    _schema = GaudiInputDataSplitter._schema.inherit_copy()
    _schema.datadict['bulksubmit'] = SimpleItem(defvalue=False,
                                                doc='determines if subjobs are split '
                                                'server side in a "bulk" submission or '
                                                'split locally and submitted individually')
    _schema.datadict['ignoremissing'] = SimpleItem(defvalue=False,
                                                   doc='Skip LFNs if they are not found '
                                                   'in the LFC. This option is only used if'
                                                   'jobs backend is Dirac')

    def _attribute_filter__set__(self, name, value):
        if name is 'sizePerJob':
            if value > 100:
                logger.warning('filesPerJob exceeded DIRAC maximum')
                logger.warning('DIRAC has a maximum dataset limit of 100.')
                logger.warning(
                    'BE AWARE!... will set it to this maximum value at submit time if backend is Dirac')
        return value

    def _create_subjob(self, job, dataset):
        logger.debug("_create_subjob")

        datatmp = []
        if isinstance(dataset, LHCbDataset):
            for i in dataset:
                if isinstance(i, DiracFile):
                    datatmp.extend(i)
                else:
                    logger.error(
                        "Unkown file-type %s, cannot perform split with file %s" % (type(i), str(i)))
                    from Ganga.Core.exceptions import GangaException
                    raise GangaException(
                        "Unkown file-type %s, cannot perform split with file %s" % (type(i), str(i)))
        elif isinstance(dataset, list):
            from Ganga.GPIDev.Base.Proxy import isType
            for i in dataset:
                if type(i) is str:
                    datatmp.append(DiracFile(lfn=i))
                elif isType(i, DiracFile()):
                    datatmp.extend(i)
                else:
                    x = GangaException(
                        "Unknown(unexpected) file object: %s" % i)
                    raise x
        else:
            logger.error("Unkown dataset type, cannot perform split here")
            from Ganga.Core.exceptions import GangaException
            raise GangaException(
                "Unkown dataset type, cannot perform split here")

        logger.debug("Creating new Job in Splitter")
        j = Job()
        j.copyFrom(stripProxy(job))
        j.splitter = None
        j.merger = None
        j.inputsandbox = []  # master added automatically
        j.inputfiles = []
        j.inputdata = LHCbDataset(files=datatmp[:],
                                  persistency=self.persistency,
                                  depth=self.depth)
        j.inputdata.XMLCatalogueSlice = self.XMLCatalogueSlice

        return j

    # returns splitter generator
    def _splitter(self, job, inputdata):

        logger.debug("_splitter")

        indata = stripProxy(copy.deepcopy(job.inputdata))

        if not job.inputdata:
            share_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),
                                      'shared',
                                      getConfig('Configuration')['user'],
                                      job.application.is_prepared.name,
                                      'inputdata',
                                      'options_data.pkl')
            if os.path.exists(share_path):
                f = open(share_path, 'r+b')
                indata = pickle.load(f)
                f.close()
            else:
                logger.error('Cannot split if no inputdata given!')
                raise SplittingError(
                    'job.inputdata is None and no inputdata found in optsfile')

        self.depth = indata.depth
        self.persistency = indata.persistency
        self.XMLCatalogueSlice = indata.XMLCatalogueSlice

        if stripProxy(job.backend).__module__.find('Dirac') > 0:
            if self.filesPerJob > 100:
                self.filesPerJob = 100  # see above warning
            logger.debug("indata: %s " % str(indata))
            outdata = DiracSplitter(indata,
                                    self.filesPerJob,
                                    self.maxFiles,
                                    self.ignoremissing)
            logger.debug("outdata: %s " % str(outdata))
            return outdata
        else:
            logger.error(
                "This Splitter HAS NOT, yet been implemented for all IGangaFile objects")
            raise NotImplementedError
            # return super(SplitFilesBySize,self)._splitter(job, indata)

    def split(self, job):
        logger.debug("split")
        if self.maxFiles == -1:
            self.maxFiles = None
        if self.bulksubmit:
            if stripProxy(job.backend).__module__.find('Dirac') > 0:
                logger.debug("Returning []")
                return []
        split_return = super(SplitFilesBySize, self).split(job)
        logger.debug("split_return: %s" % split_return)
        return split_return



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 19/02/2015
################################################################################
"""@package oaReconPlusoaAnalysis
This module is designed to run any ND280 executable accessible in the $PATH environment variable.
"""

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IPrepareApp import IPrepareApp
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config import getConfig

from Ganga.GPIDev.Lib.File import *
from Ganga.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.Core import ApplicationConfigurationError

import os, shutil, commands, re, time
from Ganga.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class oaReconPlusoaAnalysis(IApplication):
    """
    oaReconPlusoaAnalysis application running any ND280 executables.

    The required input for this module are:
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
    You can modify the output filename by replacing a string of the input filename.
    For example to replace "reco" in the input filename by "newreco" in the output of oaRecon:
        app.filenamesubstr = "reco"
        app.reconewstr = "newreco"


    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to analyze only 10 events:
        app.reco_args = ['-n',10]
    Or to use a customized parameters file:
        app.reco_args = ['-O','par_override=/home/me/myparametersfile.dat']
    """
    _schema = Schema(Version(1,1), {
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'reco_args' : SimpleItem(defvalue=[],typelist=['str','Ganga.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'anal_args' : SimpleItem(defvalue=[],typelist=['str','Ganga.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'filenamesubstr' : SimpleItem(defvalue=None,doc='This string will be substituted by recnewstr variable in the reco output filename and analnewstr variable in the oaAnalysis output filename.', typelist=['str','type(None)']),
        'reconewstr' : SimpleItem(defvalue='newreco',doc='This string will substitute filenamesubstr in the input filename to create the reco output filename.', typelist=['str']),
        'analnewstr' : SimpleItem(defvalue='validtree',doc='This string will substitute filenamesubstr in the input filename to create the oaAnalysis output filename.', typelist=['str']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        } )
    _category = 'applications'
    _name = 'oaReconPlusoaAnalysis'
    _scriptname = None
    _exportmethods = []
    _GUIPrefs = [ { 'attribute' : 'reco_args', 'widget' : 'String_List' },
                  { 'attribute' : 'anal_args', 'widget' : 'String_List' },
                  { 'attribute' : 'filenamesubstr', 'widget' : 'String' },
                  { 'attribute' : 'reconewstr', 'widget' : 'String' },
                  { 'attribute' : 'analnewstr', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'reco_args', 'widget' : 'String_List' },
                          { 'attribute' : 'anal_args', 'widget' : 'String_List' },
                          { 'attribute' : 'filenamesubstr', 'widget' : 'String' },
                          { 'attribute' : 'reconewstr', 'widget' : 'String' },
                          { 'attribute' : 'analnewstr', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    def __init__(self):
        super(oaReconPlusoaAnalysis,self).__init__()


    def configure(self,masterappconfig):
        if self.cmtsetup == None:
          raise ApplicationConfigurationError(None,'No cmt setup script given.')

        # __________ Reco first ____________
        reco_args = convertIntToStringArgs(self.reco_args)

        job = self.getJobObject()

        # Need to handle the possibility of multiple output files !
        # setup the output file
        for arg in reco_args:
          if arg == '-o':
            raise ApplicationConfigurationError(None,'Option "-o" given in reco_args. You must use the filenamesubstr and reconewstr variables instead to define an output.')

        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        if job.inputdata == None:
          raise ApplicationConfigurationError(None,'The inputdata variable is not defined.')
        fileList = job.inputdata.get_dataset_filenames()
        if len(fileList) < 1:
          raise ApplicationConfigurationError(None,'No input data file given.')

        firstFile = fileList[0].split('/')[-1]
        # Define the output
        reco_args.append('-o')
        if self.filenamesubstr == None:
          reco_outputfile = 'recoOutput.root'
        else:
          reco_outputfile = firstFile.replace(self.filenamesubstr, self.reconewstr)

        reco_args.append(reco_outputfile)
        
        # Just to define the output before the potentially long list of input files
        reco_args.extend(fileList)

        # __________ Now oaAnalysis ____________
        anal_args = convertIntToStringArgs(self.anal_args)

        job = self.getJobObject()

        # Need to handle the possibility of multiple output files !
        # setup the output file
        for arg in anal_args:
          if arg == '-o':
            raise ApplicationConfigurationError(None,'Option "-o" given in anal_args. You must use the filenamesubstr and reconewstr variables instead to define an output.')

        # Define the output
        anal_args.append('-o')
        if self.filenamesubstr == None:
          anal_outputfile = 'analOutput.root'
        else:
          anal_outputfile = firstFile.replace(self.filenamesubstr, self.analnewstr)
        anal_args.append(anal_outputfile)

        # Now add the input file
        anal_args.append(reco_outputfile)

        reco_argsStr = ' '.join(reco_args)
        anal_argsStr = ' '.join(anal_args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        script += 'RunOARecon.exe '+reco_argsStr+'\n'
        script += 'RunOAAnalysis.exe '+anal_argsStr+'\n'

        from Ganga.GPIDev.Lib.File import FileBuffer

        scriptname = 'oaReconPlusoaAnalysis.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_oaReconPlusoaAnalysis') #_Properties


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('oaReconPlusoaAnalysis','LSF', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','Local', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','PBS', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','SGE', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','Condor', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','LCG', LCGRTHandler)
allHandlers.add('oaReconPlusoaAnalysis','gLite', gLiteRTHandler)
allHandlers.add('oaReconPlusoaAnalysis','TestSubmitter', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','Interactive', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','Batch', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','Cronus', RTHandler)
allHandlers.add('oaReconPlusoaAnalysis','Remote', LCGRTHandler)
allHandlers.add('oaReconPlusoaAnalysis','CREAM', LCGRTHandler)

